#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

"""
The :mod:`gruvi.dbus` module implements a D-BUS client and server.

The implementation uses parts of the `txdbus
<https://github.com/cocagne/txdbus>`_ project. A cut down copy of txdbus,
containing only those parts needed by Gruvi, is available as ``gruvi.txdbus``.
You need this if you are providing a message handler (see below).

Both a client and a server/bus-side implementation are provided. The bus-side
implementation is very bare bones and apart from the "Hello" message it does
not implement any of the "org.freedestkop.DBus" interface. It also does not
implement any message routing. The server side is provided mostly for testing
purposes (but it could serve as the basis for a real D-BUS server).

The client side of a D-BUS connection is implemented by :class:`DbusClient` and
the server/bus-side by :class:`DbusServer`. Both implement a procedural
interface. Messages can be send using e.g. :meth:`DbusClient.send_message` or
:meth:`DbusClient.call_method`. An object-oriented interface that represents
D-BUS objects as Python objects, like the one txdbus provides, is currently not
available. The procedural interface can be used as a basis for your own
object-oriented interface though.

To receive notifications or to respond to method calls, you need to provide a
*message handler* to the client or the server constructor. The signature of the
message handler is: ``message_handler(message, protocol)``. Here, the *message*
argument is an instance of ``gruvi.txdbus.DbusMessages``, and the
*protocol* will be the :class:`DbusProtocol` instance for the current
connection.

Message handlers runs in their own fiber, which allows them to call into
switchpoints. There is one fiber for every connection.

Usage example::

  client = gruvi.DbusClient()
  client.connect('session')
  result = client.call_method('org.freedesktop.DBus', '/org/freedesktop/DBus',
                              'org.freedesktop.DBus', 'ListNames')
  for name in result[0]:
      print('Name: {}'.format(name))
"""

from __future__ import absolute_import, print_function

import os
import struct
import binascii
import codecs
import functools
import six
import pyuv

from . import compat
from .hub import switchpoint, switch_back
from .util import delegate_method
from .sync import Event
from .transports import TransportError
from .protocols import ProtocolError, MessageProtocol
from .stream import Stream
from .endpoints import Client, Server
from .address import saddr
from .vendor import txdbus

__all__ = ['DbusError', 'DbusMethodCallError', 'DbusProtocol', 'DbusClient', 'DbusServer']


class DbusError(ProtocolError):
    """Exception that is raised in case of D-BUS protocol errors."""


class DbusMethodCallError(DbusError):
    """Exception that is raised when a error reply is received for a D-BUS
    method call."""

    def __init__(self, method, reply):
        message = 'error calling {!r} method ({})'.format(method, reply.error_name)
        super(DbusMethodCallError, self).__init__(message)
        self._error = reply.error_name
        self._args = tuple(reply.body) if reply.body else ()

    @property
    def error(self):
        return self._error

    @property
    def args(self):
        return self._args


def parse_dbus_address(address):
    """Parse a D-BUS address string into a list of addresses."""
    if address == 'session':
        address = os.environ.get('DBUS_SESSION_BUS_ADDRESS')
        if not address:
            raise ValueError('$DBUS_SESSION_BUS_ADDRESS not set')
    elif address == 'system':
        address = os.environ.get('DBUS_SYSTEM_BUS_ADDRESS',
                                 'unix:path=/var/run/dbus/system_bus_socket')
    addresses = []
    for addr in address.split(';'):
        p1 = addr.find(':')
        if p1 == -1:
            raise ValueError('illegal address string: {}'.format(addr))
        kind = addr[:p1]
        args = dict((kv.split('=') for kv in addr[p1+1:].split(',')))
        if kind == 'unix':
            if 'path' in args:
                addr = args['path']
            elif 'abstract' in args:
                addr = '\0' + args['abstract']
            else:
                raise ValueError('require "path" or "abstract" for unix')
        elif kind == 'tcp':
            if 'host' not in args or 'port' not in args:
                raise ValueError('require "host" and "port" for tcp')
            addr = (args['host'], int(args['port']))
        else:
            raise ValueError('unknown transport: {}'.format(kind))
        addresses.append(addr)
    return addresses


class TxdbusAuthenticator(object):
    """A adapter to use the txdbus client and server authenticators with our
    transports and protocols."""

    # For testing, cookie_dir is set to a temporary path. Otherwise, txdbus
    # uses ~/.dbus-keyrings as specified in the spec.
    cookie_dir = None

    def __init__(self, transport, server_side, server_guid=None):
        self._transport = transport
        self._server_side = server_side
        if self._server_side:
            self._authenticator = txdbus.BusAuthenticator(server_guid)
            self._authenticator.authenticators['DBUS_COOKIE_SHA1'].keyring_dir = self.cookie_dir
        else:
            self._authenticator = txdbus.ClientAuthenticator()
            self._authenticator.cookie_dir = self.cookie_dir
        self._authenticator.beginAuthentication(self)

    def sendAuthMessage(self, message):
        # Called by the txdbus authenticators
        message = message.encode('ascii') + b'\r\n'
        self._transport.write(message)

    @property
    def _unix_creds(self):
        # Used by txdbus.BusExternalAuthenticator
        return self._transport.get_extra_info('unix_creds')

    def handleAuthMessage(self, line):
        # Called by our protocol
        self._authenticator.handleAuthMessage(line)

    def authenticationSucceeded(self):
        """Return whether the authentication succeeded."""
        return self._authenticator.authenticationSucceeded()

    def getMechanismName(self):
        """Return the authentication mechanism name."""
        if self._server_side:
            mech = self._authenticator.current_mech
            return mech.getMechanismName() if mech else None
        else:
            return getattr(self._authenticator, 'authMech', None)

    def getUserName(self):
        """Return the authenticated user name (server side)."""
        if not self._server_side:
            return
        mech = self._authenticator.current_mech
        return mech.getUserName() if mech else None

    def getGUID(self):
        """Return the GUID of the authenticated server."""
        return self._authenticator.getGUID()


def parse_dbus_header(header):
    """Parse a D-BUS header. Return the message size."""
    if six.indexbytes(header, 0) == ord('l'):
        endian = '<'
    elif six.indexbytes(header, 0) == ord('B'):
        endian = '>'
    else:
        raise ValueError('illegal endianness')
    if not 1 <= six.indexbytes(header, 1) <= 4:
        raise ValueError('illegel message type')
    if struct.unpack(endian + 'I', header[8:12])[0] == 0:
        raise ValueError('illegal serial number')
    harrlen = struct.unpack(endian + 'I', header[12:16])[0]
    padlen = (8 - harrlen) % 8
    bodylen = struct.unpack(endian + 'I', header[4:8])[0]
    return 16 + harrlen + padlen + bodylen


def new_server_guid():
    """Return a new GUID for a server."""
    return binascii.hexlify(os.urandom(16)).decode('ascii')


class DbusProtocol(MessageProtocol):
    """D-BUS Protocol."""

    # According to the D-BUS spec the max message size is 128MB. However since
    # we want to limited memory usage we are much more conservative here.
    max_message_size = 128*1024

    # Maximum size for an authentication line
    max_line_size = 1000

    _next_unique_name = 0

    S_CREDS_BYTE, S_AUTHENTICATE, S_MESSAGE_HEADER, S_MESSAGE = range(4)

    def __init__(self, message_handler=None, server_side=False, server_guid=None, timeout=None):
        super(DbusProtocol, self).__init__(message_handler, timeout=timeout)
        self._server_side = server_side
        self._name_acquired = Event()
        self._buffer = bytearray()
        self._method_calls = {}
        self._authenticator = None
        if self._server_side:
            self._server_guid = server_guid or new_server_guid()
            self._unique_name = ':{}'.format(self._next_unique_name)
            type(self)._next_unique_name += 1
        else:
            self._server_guid = None
            self._unique_name = None
        self._state = None

    @property
    def server_guid(self):
        return self._server_guid

    def connection_made(self, transport):
        # Protocol callback
        super(DbusProtocol, self).connection_made(transport)
        # The client initiates by sending a '\0' byte, as per the D-BUS spec.
        if self._server_side:
            self._state = self.S_CREDS_BYTE
        else:
            self._state = self.S_AUTHENTICATE
            self._transport.write(b'\0')
        self._writer = Stream(transport, 'w')
        self._authenticator = TxdbusAuthenticator(transport, self._server_side, self._server_guid)
        self._message_size = 0

    def connection_lost(self, exc):
        # Protocol callback
        super(DbusProtocol, self).connection_lost(exc)
        if self._error is None:
            self._error = TransportError('connection lost')
        for notify in self._method_calls.values():
            if isinstance(notify, switch_back):
                notify.throw(self._error)
        self._method_calls.clear()
        self._name_acquired.set()
        self._authenticator = None  # break cycle

    def on_creds_byte(self, byte):
        if byte != 0:
            self._error = DbusError('first byte needs to be zero')
            return False
        self._state = self.S_AUTHENTICATE
        return True

    def on_partial_auth_line(self, line):
        if len(line) > self.max_line_size:
            self._error = DbusError('auth line too long ({} bytes)'.format(len(line)))
            return False
        return True

    def on_auth_line(self, line):
        if not self.on_partial_auth_line(line):
            return False
        if line[-2:] != b'\r\n':
            self._error = DbusError('auth line does not end with \\r\\n')
            return False
        try:
            line = codecs.decode(line[:-2], 'ascii')  # codecs.decode allows memoryview
        except UnicodeDecodeError as e:
            self._error = DbusError('auth line contain non-ascii chars')
            return False
        try:
            self._authenticator.handleAuthMessage(line)
        except txdbus.DBusAuthenticationFailed as e:
            self._error = DbusError('authentication failed: {!s}'.format(e))
            return False
        if self._authenticator.authenticationSucceeded():
            if not self._server_side:
                message = txdbus.MethodCallMessage('/org/freedesktop/DBus', 'Hello',
                                    'org.freedesktop.DBus', 'org.freedesktop.DBus')
                self._transport.write(message.rawMessage)
                self._method_calls[message.serial] = self.on_hello_response
            self._state = self.S_MESSAGE_HEADER
            self._server_guid = self._authenticator.getGUID()
        return True

    def on_hello_response(self, message):
        self._unique_name = message.body[0]
        self._name_acquired.set()

    def on_message_header(self, header):
        try:
            size = parse_dbus_header(header)
        except ValueError:
            self._error = DbusError('invalid message header')
            return False
        if size > self.max_message_size:
            self._error = DbusError('message too large ({} bytes)'.format(size))
            return False
        self._message_size = size
        self._state = self.S_MESSAGE
        return True

    def on_message(self, message):
        try:
            parsed = txdbus.parseMessage(message)
        except (txdbus.MarshallingError, struct.error) as e:
            self._error = DbusError('parseMessage() error: {!s}'.format(e))
            return False
        if self._server_side and not self._name_acquired.is_set():
            if isinstance(parsed, txdbus.MethodCallMessage) \
                        and parsed.member == 'Hello' \
                        and parsed.path == '/org/freedesktop/DBus' \
                        and parsed.interface == 'org.freedesktop.DBus' \
                        and parsed.destination == 'org.freedesktop.DBus':
                response = txdbus.MethodReturnMessage(parsed.serial, signature='s',
                                                      body=[self._unique_name])
                self._name_acquired.set()
                self._transport.write(response.rawMessage)
            else:
                self._error = DbusError('Hello method not called')
                return False
        elif isinstance(parsed, (txdbus.MethodReturnMessage, txdbus.ErrorMessage)) \
                    and getattr(parsed, 'reply_serial', 0) in self._method_calls:
            notify = self._method_calls.pop(parsed.reply_serial)
            notify(parsed)
        elif self._dispatcher:
            self._queue.put_nowait(parsed)
        else:
            mtype = type(parsed).__name__[:-7].lower()
            info = ' {!r}'.format(getattr(parsed, 'member', getattr(parsed, 'error_name', '')))
            self._log.warning('no handler, ignoring inbound {}{}', mtype, info)
        self._state = self.S_MESSAGE_HEADER
        return True

    def prepend_buffer(self, buf):
        if self._buffer:
            self._buffer.extend(buf)
            buf = self._buffer
            self._buffer = bytearray()
        return memoryview(buf)

    def data_received(self, data):
        view = memoryview(data)
        offset = 0
        while offset != len(data):
            if self._state == self.S_CREDS_BYTE:
                credsbyte = six.indexbytes(view, offset)
                offset += 1
                if not self.on_creds_byte(credsbyte):
                    break
            if self._state == self.S_AUTHENTICATE:
                pos = data.find(b'\n', offset)
                if pos == -1:
                    self._buffer.extend(view[offset:])
                    self.on_partial_auth_line(self._buffer)
                    break
                line = self.prepend_buffer(view[offset:pos+1])
                offset = pos+1
                if not self.on_auth_line(line):
                    break
            if self._state == self.S_MESSAGE_HEADER:
                needbytes = 16 - len(self._buffer)
                if len(data) - offset < needbytes:
                    self._buffer.extend(view[offset:])
                    break
                header = self.prepend_buffer(view[offset:offset+needbytes])
                if not self.on_message_header(header):
                    break
                offset += len(header)
                self._buffer.extend(header)
            if self._state == self.S_MESSAGE:
                needbytes = self._message_size - len(self._buffer)
                if len(data) - offset < needbytes:
                    self._buffer.extend(view[offset:])
                    break
                message = self.prepend_buffer(view[offset:offset+needbytes])
                offset += needbytes
                if not self.on_message(message):
                    break
        self._maybe_pause_transport()
        if self._error:
            self._transport.close()
            return

    @switchpoint
    def get_unique_name(self):
        """Return the unique name of the D-BUS connection."""
        self._name_acquired.wait()
        if self._error:
            raise compat.saved_exc(self._error)
        elif self._transport is None:
            raise DbusError('not connected')
        return self._unique_name

    @switchpoint
    def send_message(self, message):
        """Send a D-BUS message.

        The *message* argument must be ``gruvi.txdbus.DbusMessage`` instance.
        """
        if not isinstance(message, txdbus.DbusMessage):
            raise TypeError('message: expecting DbusMessage instance (got {!r})',
                                type(message).__name__)
        self._name_acquired.wait()
        if self._error:
            raise compat.saved_exc(self._error)
        elif self._transport is None:
            raise DbusError('not connected')
        self._writer.write(message.rawMessage)

    @switchpoint
    def call_method(self, service, path, interface, method, signature=None,
                    args=None, no_reply=False, auto_start=False, timeout=-1):
        """Call a D-BUS method and wait for its reply.

        This method calls the D-BUS method with name *method* that resides on
        the object at bus address *service*, at path *path*, on interface
        *interface*.

        The *signature* and *args* are optional arguments that can be used to
        add parameters to the method call. The signature is a D-BUS signature
        string, while *args* must be a sequence of python types that can be
        converted into the types specified by the signature. See the `D-BUS
        specification
        <http://dbus.freedesktop.org/doc/dbus-specification.html>`_ for a
        reference on signature strings.

        The flags *no_reply* and *auto_start* control the NO_REPLY_EXPECTED and
        NO_AUTO_START flags on the D-BUS message.

        The return value is the result of the D-BUS method call. This will be a
        possibly empty sequence of values.
        """
        message = txdbus.MethodCallMessage(path, method, interface=interface,
                                destination=service, signature=signature, body=args,
                                expectReply=not no_reply, autoStart=auto_start)
        serial = message.serial
        if timeout == -1:
            timeout = self._timeout
        try:
            with switch_back(timeout) as switcher:
                self._method_calls[serial] = switcher
                self.send_message(message)
                args, _ = self._hub.switch()
        finally:
            self._method_calls.pop(serial, None)
        response = args[0]
        assert response.reply_serial == serial
        if isinstance(response, txdbus.ErrorMessage):
            raise DbusMethodCallError(method, response)
        args = tuple(response.body) if response.body else ()
        return args


class DbusClient(Client):
    """A D-BUS client."""

    def __init__(self, message_handler=None, timeout=30):
        """
        The *message_handler* argument specifies an optional message handler.

        The optional *timeout* argument specifies a default timeout for
        protocol operations in seconds.
        """
        protocol_factory = functools.partial(DbusProtocol, message_handler)
        super(DbusClient, self).__init__(protocol_factory, timeout)

    @switchpoint
    def connect(self, address='session'):
        """Connect to *address* and wait until the connection is established.

        The *address* argument must be a D-BUS server address, in the format
        described in the D-BUS specification. It may also be one of the special
        addresses ``'session'`` or ``'system'``, to connect to the D-BUS
        session and system bus, respectively.
        """
        if isinstance(address, six.string_types):
            addresses = parse_dbus_address(address)
        else:
            addresses = [address]
        for addr in addresses:
            try:
                super(DbusClient, self).connect(addr)
            except pyuv.error.UVError:
                continue
            break
        else:
            raise DbusError('could not connect to any address')
        # Wait for authentication to complete
        self.get_unique_name()

    protocol = Client.protocol

    delegate_method(protocol, DbusProtocol.get_unique_name)
    delegate_method(protocol, DbusProtocol.send_message)
    delegate_method(protocol, DbusProtocol.call_method)


class DbusServer(Server):
    """A D-BUS server."""

    def __init__(self, message_handler, timeout=30):
        """
        The *message_handler* argument specifies the message handler.

        The optional *timeout* argument specifies a default timeout for
        protocol operations in seconds.
        """
        protocol_factory = functools.partial(DbusProtocol, message_handler,
                                             server_side=True)
        super(DbusServer, self).__init__(protocol_factory, timeout)

    @switchpoint
    def listen(self, address='session'):
        """Start listening on *address* for new connection.

        The *address* argument must be a D-BUS server address, in the format
        described in the D-BUS specification. It may also be one of the special
        addresses ``'session'`` or ``'system'``, to connect to the D-BUS
        session and system bus, respectively.
        """
        if isinstance(address, six.string_types):
            addresses = parse_dbus_address(address)
        else:
            addresses = [address]
        for addr in addresses:
            try:
                super(DbusServer, self).listen(addr)
            except pyuv.error.UVError:
                self._log.error('skipping address {}', saddr(addr))



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import os
import sys
import logging
import threading
import fibers
import six

from . import util

__all__ = ['get_logger']

# Add a new level: TRACE.
logging.TRACE = 5
assert logging.NOTSET < logging.TRACE < logging.DEBUG
logging.addLevelName('TRACE', logging.TRACE)

_logger_name = 'gruvi'
_logger_dict = {}

# The logging module documents this slight hack to disable finding caller
# information (via sys._getframe()) for every logging call. In our logger we
# only get logging information if needed (at the DEBUG level or higher), so we
# can disable collecting it for every call.
logging._srcfile = None


def get_logger(context=None, name=None):
    """Return a logger for *context*.

    Return a :class:`ContextLogger` instance. The instance implements the
    standard library's :class:`logging.Logger` interface.
    """
    # Many class instances have their own logger. Share them to save memory if
    # possible, i.e. when *context* is not set.
    if name is None:
        name = _logger_name
    if context is None and name in _logger_dict:
        return _logger_dict[name]
    if context is not None and not isinstance(context, six.string_types):
        context = util.objref(context)
    logger = logging.getLogger(name)
    logger = ContextLogger(logger, context)
    if context is None:
        _logger_dict[name] = logger
    return logger


class ContextLogger(object):
    """A logger adapter that prepends a context string to log messages.

    It also supports passing arguments via '{}' format operations.
    """

    __slots__ = ('_logger', '_context')

    # This is not based on logging.LoggingAdapter because the 2.x and 3.x
    # implementations differ quite a bit, which means we would need to
    # reimplement almost the entire thing anyway.

    def __init__(self, logger, context=None):
        self._logger = logger
        self._context = context or ''

    @property
    def context(self):
        """Return the logging context."""
        return self._context

    def thread_info(self):
        """Return a string identifying the current thread and fiber."""
        tid = threading.current_thread().name
        if tid == 'MainThread':
            tid = 'Main'
        current = fibers.current()
        fid = getattr(current, 'name', util.objref(current)) if current.parent else 'Root'
        return '{}/{}'.format(tid, fid)

    def frame_info(self):
        """Return a string identifying the current frame."""
        if not self._logger.isEnabledFor(logging.DEBUG):
            return ''
        f = sys._getframe(3)
        fname = os.path.split(f.f_code.co_filename)[1]
        return '{}:{}'.format(fname, f.f_lineno)

    def log(self, level, msg, *args, **kwargs):
        if not self._logger.isEnabledFor(level):
            return
        prefix = '{}|{}|{}'.format(self.thread_info(), self.context or '-', self.frame_info())
        if args:
            msg = msg.format(*args)
        msg = '[{}] {}'.format(prefix, msg)
        self._logger._log(level, msg, (), **kwargs)

    def trace(self, msg, *args, **kwargs):
        self.log(logging.TRACE, msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        self.log(logging.DEBUG, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.log(logging.INFO, msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.log(logging.WARNING, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.log(logging.ERROR, msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.log(logging.CRITICAL, msg, *args, **kwargs)

    def exception(self, msg, *args, **kwargs):
        kwargs['exc_info'] = True
        self.log(logging.ERROR, msg, *args, **kwargs)



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

__all__ = ['Error', 'Timeout', 'Cancelled']


class Error(Exception):
    """Base class for Gruvi exceptions."""

class Timeout(Error):
    """A timeout has occurred."""

class Cancelled(BaseException):
    """A fiber or calback was cancelled."""



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import os
import sys
import shutil
import socket
import errno
import tempfile
import logging
import subprocess
import pkg_resources
import unittest
import ssl
import six

import gruvi
from gruvi.util import split_cap_words
from gruvi.ssl import create_ssl_context
from gruvi.sync import Event

__all__ = []


def setup_logging():
    """Configure a logger to output to stdout."""
    logger = logging.getLogger()
    if logger.handlers:
        return
    handler = logging.StreamHandler(sys.stdout)
    debug = int(os.environ.get('DEBUG', '0'))
    verbose = int(os.environ.get('VERBOSE', '5' if debug else '2'))
    # Smarty-pants way to say 0 = no logs (60), 1 = CRITICAL (50), ... 6 = TRACE (5)
    level = max(5, 10 * (6 - verbose))
    logger.setLevel(level)
    template = '%(levelname)s %(message)s'
    handler.setFormatter(logging.Formatter(template))
    logger.addHandler(handler)


def create_ssl_certificate(fname):
    """Create a new SSL private key and self-signed certificate, and store
    them both in the file *fname*."""
    try:
        openssl = subprocess.Popen(['openssl', 'req', '-new',
                        '-newkey', 'rsa:1024', '-x509', '-subj', '/CN=test/',
                        '-days', '365', '-nodes', '-batch',
                        '-out', fname, '-keyout', fname],
                        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError:
        sys.stderr.write('Error: openssl not found. SSL tests disabled.\n')
        return
    stdout, stderr = openssl.communicate()
    if openssl.returncode:
        sys.stderr.write('Error: key generation failed\n')
        sys.stderr.write('openssl stdout: {0}\n'.format(stdout))
        sys.stderr.write('openssl stderr: {0}\n'.format(stderr))


def create_cmd_wrappers(bindir):
    """On Windows, Create executable file wrappers for our utilities in tests/bin."""
    # This is relevant on Windows only. On Unix our utilities can be executed
    # by uv_spawn() directly.
    #
    # On Windows, a simple solution could be to create .bat file wrappers.
    # However that doesn't work because uv_spawn() uses CreateProcess() which
    # only supports .exe and .com files.
    #
    # The solution is to create little .exe wrapper for each program.
    # Fortunately this is easy. Setuptools contains such a wrapper as a package
    # resource. We need to copy it, and create a basename-script.py wrapper.
    if not sys.platform.startswith('win'):
        return
    shebang = '#!{0}\r\n'.format(sys.executable)
    wrapper = None
    for fname in os.listdir(bindir):
        if '.' in fname:
            continue
        absname = os.path.join(bindir, fname)
        scriptname = absname + '-script.py'
        exename = absname + '.exe'
        if os.access(scriptname, os.R_OK) and os.access(exename, os.X_OK):
            continue
        with open(absname) as fin:
            lines = [line.rstrip() + '\r\n' for line in fin.readlines()]
            lines[0] = shebang
        with open(scriptname, 'w') as fout:
            fout.writelines(lines)
        if wrapper is None:
            wrapper = pkg_resources.resource_string('setuptools', 'cli.exe')
        with open(exename, 'wb') as fout:
            fout.write(wrapper)


def sizeof(obj, exclude=None):
    """Return the size in bytes of *obj*."""
    if obj is None or obj is False or obj is True:
        return 0
    size = sys.getsizeof(obj)
    if hasattr(obj, '__dict__'):
        size += sys.getsizeof(obj.__dict__)
        for key, value in obj.__dict__.items():
            if exclude is not None and key in exclude:
                continue
            s = sizeof(key)
            s += sizeof(value, exclude)
            # print('{}.{}: {}'.format(type(obj).__name__, key, s))
            size += s
    elif hasattr(obj, '__slots__'):
        for key in obj.__slots__:
            if hasattr(obj, key):
                if exclude is not None and key in exclude:
                    continue
                s = sizeof(getattr(obj, key), exclude)
                # print('{}.{}: {}'.format(type(obj).__name__, key, s))
                size += s
    return size


def socketpair(family=socket.AF_INET, type=socket.SOCK_STREAM, proto=0):
    """Emulate the Unix socketpair() syscall by connecting an AF_INET socket."""
    # This is useful on platforms like Windows that don't have one.
    # We create a connected TCP socket. Note the trick with setblocking(0)
    # that prevents us from having to create a thread.
    lsock = socket.socket(family, type, proto)
    lsock.bind(('localhost', 0))
    lsock.listen(1)
    addr, port = lsock.getsockname()
    csock = socket.socket(family, type, proto)
    csock.setblocking(False)
    try:
        csock.connect((addr, port))
    except socket.error as e:
        if e.errno not in (errno.EAGAIN, errno.EWOULDBLOCK, errno.EINPROGRESS):
            lsock.close()
            csock.close()
            raise
    ssock, _ = lsock.accept()
    csock.setblocking(True)
    lsock.close()
    return (ssock, csock)


class TestCase(unittest.TestCase):
    """Base class for test suites."""

    test_prefix = 'test'

    @classmethod
    def setUpClass(cls):
        setup_logging()
        cls.testdir = os.path.abspath(os.path.split(__file__)[0])
        cls.topdir = os.path.split(cls.testdir)[0]
        os.chdir(cls.testdir)
        certname = 'testcert.pem'
        if not os.access(certname, os.R_OK):
            create_ssl_certificate(certname)
        cls.certname = certname
        bindir = os.path.join(cls.testdir, 'bin')
        path = os.environ.get('PATH', '')
        if bindir not in path:
            create_cmd_wrappers(bindir)
            os.environ['PATH'] = os.pathsep.join([bindir, path])

    def setUp(self):
        self._tmpindex = 1
        self.__tmpdir = os.path.realpath(tempfile.mkdtemp('gruvi-test'))
        self.__tmpinode = os.stat(self.__tmpdir).st_ino

    def tearDown(self):
        # Some paranoia checks to make me feel better before calling
        # shutil.rmtree()..
        assert '/..' not in self.__tmpdir and '\\..' not in self.__tmpdir
        assert os.stat(self.__tmpdir).st_ino == self.__tmpinode
        try:
            shutil.rmtree(self.__tmpdir)
        except OSError:
            # On Windows a WindowsError is raised when files are
            # still open (WindowsError inherits from OSError).
            pass
        self.__tmpdir = None
        self.__tmpinode = None
        # Check that no active handles remain. This would mess with other tests.
        hub = gruvi.get_hub()
        active = []
        for handle in hub.loop.handles:
            if not handle.closed and not getattr(handle, '_system_handle', False):
                active.append(handle)
        for handle in active:
            print('closing active handle {0!r}'.format(handle))
            handle.close()
        if active:
            raise RuntimeError('test leaked {0} active handles'.format(len(active)))

    @classmethod
    def setup_loader(cls):
        unittest.TestLoader.testMethodPrefix = cls.test_prefix

    @property
    def tempdir(self):
        return self.__tmpdir

    def tempname(self, name=None):
        if name is None:
            name = 'tmpfile-{0}'.format(self._tmpindex)
            self._tmpindex += 1
        return os.path.join(self.__tmpdir, name)

    def pipename(self, name=None, abstract=False):
        if name is None:
            name = 'tmppipe-{0}'.format(self._tmpindex)
            self._tmpindex += 1
        if sys.platform.startswith('win'):
            return r'\\.\pipe\{0}-{1}'.format(name, os.getpid())
        else:
            prefix = '\x00' if sys.platform.startswith('linux') and abstract else ''
            return prefix + self.tempname(name)

    def get_ssl_context(self):
        context = create_ssl_context(certfile=self.certname, keyfile=self.certname)
        if hasattr(context, 'check_hostname'):
            context.check_hostname = None  # Python 3.4+
        context.verify_mode = ssl.CERT_NONE
        return context

    def assertRaises(self, exc, func, *args, **kwargs):
        # Like unittest.assertRaises, but returns the exception.
        try:
            func(*args, **kwargs)
        except exc as e:
            exc = e
        except Exception as e:
            self.fail('Wrong exception raised: {0!s}'.format(e))
        else:
            self.fail('Exception not raised: {0!s}'.format(exc))
        return exc


class UnitTest(TestCase):
    """Base class for unit tests."""


class PerformanceTest(TestCase):
    """Base class for performance tests."""

    results_name = 'performance.txt'
    test_prefix = 'perf'

    def add_result(self, result, params={}, name=None):
        """Add a performance test result."""
        if name is None:
            frame = sys._getframe(1)
            clsname = frame.f_locals.get('self', '').__class__.__name__
            methname = frame.f_code.co_name
            names = split_cap_words(clsname)
            name = '{0}_{1}'.format(''.join(names[1:]), methname[len(self.test_prefix)+1:]).lower()
        if params is not None:
            params = ','.join(['{0}={1}'.format(k, params[k]) for k in params])
        with open(self.results_name, 'a') as fout:
            fout.write('{0:<32s} {1:<16.2f} {2:s}\n'.format(name, result, params))

    @classmethod
    def start_new_results(cls):
        try:
            os.unlink(cls.results_name)
        except OSError:
            pass


class MemoryTest(PerformanceTest):
    """Special case of a performance test that writes to memory.txt."""

    results_name = 'memory.txt'
    test_prefix = 'mem'


class MockTransport(object):
    """A mock transport.

    All writes are redirected to a BytesIO instance.
    """

    default_write_buffer = 65536

    def __init__(self, mode='rw'):
        self._mode = mode
        self._readable = 'r' in mode
        self._writable = 'w' in mode
        self._protocol = None
        self._error = None
        self._reading = False
        self._writing = False
        self._can_write = Event()
        self._closed = Event()
        self._write_buffer_high = self.default_write_buffer
        self._write_buffer_low = self.default_write_buffer // 2
        self.buffer = six.BytesIO()
        self.eof = False

    def start(self, protocol):
        self._protocol = protocol
        self._protocol.connection_made(self)
        if self._readable:
            self.resume_reading()
        if self._writable:
            self._writing = True
            self._can_write.set()

    def get_write_buffer_size(self):
        return len(self.buffer.getvalue())

    def get_write_buffer_limits(self):
        return self._write_buffer_high, self._write_buffer_low

    def set_write_buffer_limits(self, high=None, low=None):
        if high is None:
            high = self._write_buffer_size
        if low is None:
            low = high // 2
        if low > high:
            low = high
        self._write_buffer_high = high
        self._write_buffer_low = low

    def drain(self):
        self.buffer.seek(0)
        self.buffer.truncate()
        self._can_write.set()
        self._protocol.resume_writing()

    def get_extra_info(self, name, default=None):
        if name == 'unix_creds':
            if hasattr(socket, 'SO_PEERCRED'):
                return (os.getpid(), os.getuid(), os.getgid())
            return default
        else:
            return default

    def pause_reading(self):
        self._reading = False

    def resume_reading(self):
        self._reading = True

    def write(self, buf):
        self.buffer.write(buf)
        if self.get_write_buffer_size() > self.get_write_buffer_limits()[0]:
            self._can_write.clear()
            self._writing = False
            self._protocol.pause_writing()

    def writelines(self, seq):
        for line in seq:
            self.write(line)

    def write_eof(self):
        self.eof = True

    def can_write_eof(self):
        return True

    def close(self):
        self._closed.set()
        self._protocol.connection_lost(None)

    def abort(self):
        self._closed.set()
        self._protocol.connection_lost(None)



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import time
import unittest

from gruvi.callbacks import dllist, Node
from support import PerformanceTest


class PerfDllist(PerformanceTest):

    def perf_insert_throughput(self):
        # Measure the number inserts we can do per second
        t0 = t1 = time.time()
        count = 0
        batch = 1000
        dll = dllist()
        value = 'foo'
        while t1 - t0 < 0.2:
            for i in range(batch):
                dll.insert(Node(value))
            count += batch
            t1 = time.time()
        speed = count / (t1 - t0)
        self.add_result(speed)

if __name__ == '__main__':
    unittest.defaultTestLoader.testMethodPrefix = 'perf'
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import os
import sys
import gc
import time
import signal
import inspect
import threading
import weakref
import unittest
from unittest import SkipTest

import gruvi
from support import UnitTest


class TestHub(UnitTest):

    def test_switchpoint(self):
        # Wrapping a function in @switchpoint should return function with
        # proper docs and a proper signature. Calling it should call the
        # underlying function.
        def func(foo, bar, baz=None, *args, **kwargs):
            """Foo bar."""
            return (foo, bar, baz, args, kwargs)
        wrapper = gruvi.switchpoint(func)
        self.assertTrue(wrapper.__switchpoint__)
        while hasattr(wrapper, '__wrapped__'):
            wrapper = wrapper.__wrapped__
        self.assertEqual(wrapper.__name__, 'func')
        argspec = inspect.getargspec(wrapper)
        self.assertEqual(argspec.args, ['foo', 'bar', 'baz'])
        self.assertEqual(argspec.varargs, 'args')
        self.assertEqual(argspec.keywords, 'kwargs')
        self.assertEqual(argspec.defaults, (None,))
        result = wrapper(1, 2, qux='foo')
        self.assertEqual(result, (1, 2, None, (), {'qux': 'foo'}))
        result = wrapper(1, 2, 3, 4, qux='foo')
        self.assertEqual(result, (1, 2, 3, (4,), {'qux': 'foo'}))

    def test_sigint(self):
        # The Hub should exit on CTRL-C (SIGINT).
        # On windows, sending SIGINT kills all processes attached to a console,
        # including the test driver.
        if sys.platform.startswith('win'):
            raise SkipTest('test skipped on Windows')
        def send_sigint():
            time.sleep(0.01)
            os.kill(os.getpid(), signal.SIGINT)
        t1 = threading.Thread(target=send_sigint)
        t1.start()
        hub = gruvi.get_hub()
        hub.switch()
        self.assertFalse(hub.is_alive())
        t1.join()

    def test_get_new_hub(self):
        # When the hub is closed, get_hub() should return a new one.
        hub = gruvi.get_hub()
        hub.close()
        hub.switch()
        hub2 = gruvi.get_hub()
        self.assertIsNot(hub, hub2)

    def test_close(self):
        # Calling hub.close() should cause the hub to exit.
        hub = gruvi.Hub()
        hub.run_callback(hub.close)
        self.assertIsNone(hub.switch())
        self.assertFalse(hub.is_alive())

    def test_close_from_thread(self):
        # Calling hub.close() from a thread should cause the hub to exit.
        def close_hub():
            time.sleep(0.01)
            hub.close()
        t1 = threading.Thread(target=close_hub)
        t1.start()
        hub = gruvi.Hub()
        self.assertIsNone(hub.switch())
        self.assertFalse(hub.is_alive())
        t1.join()

    def test_cleanup(self):
        # After we close() a Hub, it should be garbage collectable, including
        # its event loop.
        hub = gruvi.Hub()
        gruvi.sleep(0)
        ref1 = weakref.ref(hub)
        ref2 = weakref.ref(hub.loop)
        hub.close()
        hub.switch()
        del hub
        gc.collect()
        self.assertIsNone(ref1())
        self.assertIsNone(ref2())

    def test_thread_close(self):
        # If we close() a Hub in a thread, it should be garbage collectable.
        refs = []
        def thread_sleep():
            hub = gruvi.get_hub()
            gruvi.sleep(0)
            refs.append(weakref.ref(hub))
            refs.append(weakref.ref(hub.loop))
            hub.close()
            hub.switch()
        t1 = threading.Thread(target=thread_sleep)
        refs.append(weakref.ref(t1))
        t1.start(); t1.join(); del t1
        gruvi.sleep(0.1)  # Give the thread time to exit.
        gc.collect()
        self.assertIsNone(refs[0]())
        self.assertIsNone(refs[1]())
        self.assertIsNone(refs[2]())

    def test_many_hubs(self):
        # Create and close a Hub in many threads. If the hub does not garbage
        # collect, then this will run out of file descriptors.
        for i in range(100):
            hub = gruvi.Hub()
            with gruvi.switch_back(timeout=0, hub=hub):
                self.assertRaises(gruvi.Timeout, hub.switch)
            hub.close()
            hub.switch()
            gc.collect()

    def test_callback_order(self):
        # Callbacks run with run_callback() should be run in the order they
        # were added.
        hub = gruvi.Hub()
        result = []
        for i in range(100):
            hub.run_callback(result.append, i)
        hub.close()
        hub.switch()
        self.assertEqual(len(result), 100)
        self.assertEqual(result, list(range(100)))

    def test_sleep(self):
        # Test that sleep() works
        hub = gruvi.get_hub()
        # use hub.now() as there's slight rounding errors with time.time()
        t0 = hub.loop.now()
        gruvi.sleep(0.01)
        t1 = hub.loop.now()
        self.assertGreaterEqual(t1-t0, 10)
        t0 = hub.loop.now()
        gruvi.sleep(0.1)
        t1 = hub.loop.now()
        self.assertGreaterEqual(t1-t0, 100)


class TestAssertNoSwitchpoints(UnitTest):

    def test_assert_no_switchpoints(self):
        # Calling a switchpoint while in an assert_no_switchpoints block should
        # raise an AssertionError.
        with gruvi.assert_no_switchpoints():
            self.assertRaises(AssertionError, gruvi.sleep, 0)


class TestSwitchBack(UnitTest):

    def test_call(self):
        # Calling a switch_back instance should cause hub.switch() to return
        # with an (args, kwargs) tuple.
        hub = gruvi.get_hub()
        with gruvi.switch_back() as switcher:
            hub.run_callback(lambda: switcher('foo', bar='baz'))
            self.assertEqual(hub.switch(), (('foo',), {'bar': 'baz'}))

    def test_timeout(self):
        # A timeout in a switch_back instance should cause hub.switch() to
        # raise a Timeout exception.
        hub = gruvi.get_hub()
        with gruvi.switch_back(0.01):
            self.assertRaises(gruvi.Timeout, hub.switch)

    def test_throw(self):
        # An exception thrown with the throw() method of a switch_back instance
        # should cause hub.switch to raise that exception.
        hub = gruvi.get_hub()
        with gruvi.switch_back() as switcher:
            hub.run_callback(lambda: switcher.throw(ValueError('foo')))
            exc = self.assertRaises(ValueError, hub.switch)
            self.assertEqual(exc.args[0], 'foo')


if __name__ == '__main__':
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

"""
The :mod:`gruvi.http` module implements a HTTP client and server.

The client and server are relatively complete implementations of the HTTP
protocol. Some of the supported features are keepalive, pipelining, chunked
transfers and trailers.

This implementation supports both HTTP/1.0 and HTTP/1.1. The default for the
client is 1.1, and the server will respond with the same version as the client.

Connections are kept alive by default. This means that you need to make sure
you close connections when they are no longer needed, by calling the
appropriate :meth:`~gruvi.Endpoint.close` method.

It is important to clarify how the API exposed by this module uses text and
binary data. Data that is read from or written to the HTTP header, such as the
version string, method, and headers, are text strings (``str`` on Python 3,
``str`` or ``unicode`` on Python 2). However, if the string type is unicode
aware (``str`` on Python 3, ``unicode`` on Python 2), you must make sure that
it only contains code points that are part of ISO-8859-1, which is the default
encoding specified in :rfc:`2606`. Data that is read from or written to HTTP
bodies is always binary. This is done in alignment with the WSGI spec that
requires this.

This module provides a number of APIs. Client-side there is one:

* A :class:`gruvi.Client` based API. You will use :meth:`~HttpClient.connect`
  to connect to a server, and then use :meth:`~HttpClient.request` and
  :meth:`~HttpClient.getresponse` to interact with it.

The following server-side APIs are available:

* A :class:`gruvi.Server` based API. Incoming HTTP messages are passed to a
  message handler that needs to take care of all aspects of HTTP other than
  parsing.
* A WSGI API, as described in :pep:`333`.

The server-side API is selected through the *adapter* argument to
:class:`HttpServer` constructor. The default adapter is :class:`WsgiAdapter`,
which implements the WSGI protocol. To use the raw server interface, pass the
identity function (``lambda x: x``).
"""

from __future__ import absolute_import, print_function

import re
import time
import functools
import six
from collections import namedtuple

from . import logging, compat
from .hub import switchpoint
from .util import delegate_method, docfrom
from .protocols import MessageProtocol, ProtocolError
from .stream import Stream
from .endpoints import Client, Server
from .http_ffi import lib, ffi

from six.moves import http_client

__all__ = ['HttpError', 'ParsedUrl', 'parse_url', 'HttpMessage', 'HttpRequest',
           'HttpProtocol', 'WsgiAdapter', 'HttpClient', 'HttpServer']


#: Constant indicating a HTTP request.
REQUEST = lib.HTTP_REQUEST

#: Constant indicating a HTTP response.
RESPONSE = lib.HTTP_RESPONSE

# Export some definitions from  http.client.
for name in dir(http_client):
    value = getattr(http_client, name)
    if not name.isupper() or not isinstance(value, int):
        continue
    if value in http_client.responses:
        globals()[name] = value

HTTP_PORT = http_client.HTTP_PORT
HTTPS_PORT = http_client.HTTPS_PORT
responses = http_client.responses


# The "Hop by Hop" headers as defined in RFC 2616. These may not be set by the
# HTTP handler.
hop_by_hop = frozenset(('connection', 'keep-alive', 'proxy-authenticate',
                        'proxy-authorization', 'te', 'trailers',
                        'transfer-encoding', 'upgrade'))


# Keep a cache of http-parser's HTTP methods numbers -> method strings
_http_methods = {}

for i in range(100):
    method = ffi.string(lib.http_method_str(i)).decode('ascii')
    if not method.isupper():
        break
    _http_methods[i] = method


class HttpError(ProtocolError):
    """Exception that is raised in case of HTTP protocol errors."""


# Header parsing

# RFC 2626 section 2.2 grammar definitions:
re_ws = re.compile('([ \t]+)')
re_token = re.compile('([!#$%&\'*+\-.0-9A-Z^_`a-z|~]+)')
re_qstring = re.compile(r'"(([\t !\x23-\x5b\x5d-\xff]|\\[\x00-\x7f])*)"')
re_qpair = re.compile(r'\\([\x00-\x7f])')
re_qvalue = re.compile('[qQ]=(1(\.0{0,3})?|0(\.[0-9]{0,3})?)')


def lookahead(buf, pos):
    """Return the next char at the current buffer position."""
    if pos >= len(buf):
        return None
    return buf[pos]

def accept_ws(buf, pos):
    """Skip whitespace at the current buffer position."""
    match = re_ws.match(buf, pos)
    if not match:
        return None, pos
    return buf[match.start(0):match.end(0)], match.end(0)

def accept_lit(char, buf, pos):
    """Accept a literal character at the current buffer position."""
    if pos >= len(buf) or buf[pos] != char:
        return None, pos
    return char, pos+1

def expect_lit(char, buf, pos):
    """Expect a literal character at the current buffer position."""
    if pos >= len(buf) or buf[pos] != char:
        return None, len(buf)
    return char, pos+1

def accept_re(regexp, buf, pos):
    """Accept a regular expression at the current buffer position."""
    match = regexp.match(buf, pos)
    if not match:
        return None, pos
    return buf[match.start(1):match.end(1)], match.end(0)

def expect_re(regexp, buf, pos):
    """Require a regular expression at the current buffer position."""
    match = regexp.match(buf, pos)
    if not match:
        return None, len(buf)
    return buf[match.start(1):match.end(1)], match.end(0)


def parse_content_type(header):
    """Parse the "Content-Type" header."""
    typ = subtyp = None; options = {}
    typ, pos = expect_re(re_token, header, 0)
    _, pos = expect_lit('/', header, pos)
    subtyp, pos = expect_re(re_token, header, pos)
    ctype = header[:pos] if subtyp else ''
    while pos < len(header):
        _, pos = accept_ws(header, pos)
        _, pos = expect_lit(';', header, pos)
        _, pos = accept_ws(header, pos)
        name, pos = expect_re(re_token, header, pos)
        _, pos = expect_lit('=', header, pos)
        char = lookahead(header, pos)
        if char == '"':
            value, pos = expect_re(re_qstring, header, pos)
            value = re_qpair.sub('\\1', value)
        elif char:
            value, pos = expect_re(re_token, header, pos)
        if name and value is not None:
            options[name] = value
    return ctype, options

def parse_te(header):
    """Parse the "TE" header."""
    pos = 0
    names = []
    while pos < len(header):
        name, pos = expect_re(re_token, header, pos)
        _, pos = accept_ws(header, pos)
        _, pos = accept_lit(';', header, pos)
        _, pos = accept_ws(header, pos)
        qvalue, pos = accept_re(re_qvalue, header, pos)
        if name:
            names.append((name, qvalue))
        _, pos = accept_ws(header, pos)
        _, pos = expect_lit(',', header, pos)
        _, pos = accept_ws(header, pos)
    return names

def parse_trailer(header):
    """Parse the "Trailer" header."""
    pos = 0
    names = []
    while pos < len(header):
        name, pos = expect_re(re_token, header, pos)
        if name:
            names.append(name)
        _, pos = accept_ws(header, pos)
        _, pos = expect_lit(',', header, pos)
        _, pos = accept_ws(header, pos)
    return names


weekdays = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
           'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
rfc1123_fmt = '%a, %d %b %Y %H:%M:%S GMT'

def rfc1123_date(timestamp=None):
    """Create a RFC1123 style Date header for *timestamp*."""
    if timestamp is None:
        timestamp = int(time.time())
    # The time stamp must be GMT, and cannot be localized.
    tm = time.gmtime(timestamp)
    s = rfc1123_fmt.replace('%a', weekdays[tm.tm_wday]) \
                   .replace('%b', months[tm.tm_mon-1])
    return time.strftime(s, tm)


# URL parser (using http-parser). Unlike Python's urlsplit() this doesn't do
# relatively URLs. This is a benefit IMHO, because that allows us to recognize
# absolute URLs with a missing schema ('www.example.com') and not mistaken them
# for the path component of a relative URL. To parse relative URLs you need to
# first turn them into origin-form or absolute-form.

default_ports = {'http': 80, 'https': 443, 'ws': 80, 'wss': 443}
ssl_protocols = frozenset(('https', 'wss'))
url_field_names = ('scheme', 'host', 'path', 'query', 'fragment', 'port', 'userinfo')
url_field_indices = tuple((getattr(lib, 'UF_{}'.format(name.replace('scheme', 'schema')).upper())
                           for name in url_field_names))

class ParsedUrl(namedtuple('_ParsedUrl', url_field_names)):
    """A :func:`~collections.namedtuple` with the following fields:
    ``scheme``, ``host``, ``port``, ``path``, ``query``, ``fragment`` and
    ``userinfo``.

    In addition to the tuple fields the following properties are defined:
    """

    __slots__ = ()

    @classmethod
    def from_parser(cls, parser, url):
        values = []
        for ix in url_field_indices:
            if parser.field_set & (1 << ix):
                fd = parser.field_data[ix]
                values.append(url[fd.off:fd.off+fd.len])
            else:
                values.append('')
        return cls(*values)

    @property
    def addr(self):
        """Address tuple that can be used with :func:`~gruvi.create_connection`."""
        port = self.port
        if port:
            port = int(port)
        else:
            port = default_ports.get(self.scheme or 'http')
        return (self.host, port)

    @property
    def ssl(self):
        """Whether the scheme requires SSL/TLS."""
        return self.scheme in ssl_protocols

    @property
    def target(self):
        """The "target" i.e. local part of the URL, consisting of the path and query."""
        target = self.path or '/'
        if self.query:
            target = '{}?{}'.format(target, self.query)
        return target

ParsedUrl.__new__.__doc__ = ''
ParsedUrl.__new__.__defaults__ = ('',) * len(url_field_names)


def parse_url(url, default_scheme='http', is_connect=False):
    """Parse an URL and return its components.

    The *default_scheme* argument specifies the scheme in case URL is
    an otherwise valid absolute URL but with a missing scheme.

    The *is_connect* argument must be set to ``True`` if the URL was requested
    with the HTTP CONNECT method. These URLs have a different form and need to
    be parsed differently.

    The result is a :class:`ParsedUrl` containing the URL components.
    """
    # If this is not in origin-form, authority-form or asterisk-form and no
    # scheme is present, assume it's in absolute-form with a missing scheme.
    # See RFC7230 section 5.3.
    if url[:1] not in '*/' and not is_connect and '://' not in url:
        url = '{}://{}'.format(default_scheme, url)
    burl = s2b(url)
    parser = ffi.new('struct http_parser_url *')
    lib.http_parser_url_init(parser)
    res = lib.http_parser_parse_url(ffi.from_buffer(burl), len(burl), is_connect, parser)
    if res != 0:
        raise ValueError('invalid URL')
    parsed = ParsedUrl.from_parser(parser, url)
    return parsed


# String conversions. Note that ISO-8859-1 is the default encoding for HTTP
# headers.

def s2b(s):
    """Convert a string *s* to bytes in the ISO-8859-1 encoding."""
    if type(s) is not bytes:
        s = s.encode('iso-8859-1')
    return s

def ba2s(ba):
    """Convert a byte-array to a "str" type."""
    if six.PY3:
        return ba.decode('iso-8859-1')
    else:
        return bytes(ba)

def cd2s(cd):
    """Convert a cffi cdata('char *') to a str."""
    s = ffi.string(cd)
    if six.PY3:
        s = s.decode('iso-8859-1')
    return s


def get_header(headers, name, default=None):
    """Return the value of header *name*.

    The *headers* argument must be a list of ``(name, value)`` tuples. If the
    header is found its associated value is returned, otherwise *default* is
    returned. Header names are matched case insensitively.
    """
    name = name.lower()
    for header in headers:
        if header[0].lower() == name:
            return header[1]
    return default

def remove_headers(headers, name):
    """Remove all headers with name *name*.

    The list is modified in-place and the updated list is returned.
    """
    i = 0
    name = name.lower()
    for j in range(len(headers)):
        if headers[j][0].lower() != name:
            if i != j:
                headers[i] = headers[j]
            i += 1
    del headers[i:]
    return headers


def create_chunk(buf):
    """Create a chunk for the HTTP "chunked" transfer encoding."""
    chunk = bytearray()
    chunk.extend(s2b('{:X}\r\n'.format(len(buf))))
    chunk.extend(s2b(buf))
    chunk.extend(b'\r\n')
    return chunk


def create_chunked_body_end(trailers=None):
    """Create the ending that terminates a chunked body."""
    ending = bytearray()
    ending.extend(b'0\r\n')
    if trailers:
        for name, value in trailers:
            ending.extend(s2b('{}: {}\r\n'.format(name, value)))
    ending.extend(b'\r\n')
    return ending


def create_request(version, method, url, headers):
    """Create a HTTP request header."""
    # According to my measurements using b''.join is faster that constructing a
    # bytearray.
    message = []
    message.append(s2b('{} {} HTTP/{}\r\n'.format(method, url, version)))
    for name, value in headers:
        message.append(s2b('{}: {}\r\n'.format(name, value)))
    message.append(b'\r\n')
    return b''.join(message)


def create_response(version, status, headers):
    """Create a HTTP response header."""
    message = []
    message.append(s2b('HTTP/{} {}\r\n'.format(version, status)))
    for name, value in headers:
        message.append(s2b('{}: {}\r\n'.format(name, value)))
    message.append(b'\r\n')
    return b''.join(message)


class HttpMessage(object):
    """HTTP message.

    Instances of this class are returned by :meth:`HttpClient.getresponse` and
    passed as an argument to :class:`HttpServer` message handlers.
    """

    def __init__(self):
        self._message_type = None
        self._version = None
        self._status_code = None
        self._method = None
        self._url = None
        self._parsed_url = None
        self._headers = []
        self._charset = None
        self._body = None
        self._should_keep_alive = None

    @property
    def message_type(self):
        """The message type, either :data:`REQUEST` or :data:`RESPONSE`."""
        return self._message_type

    @property
    def version(self):
        """The HTTP version as a string, either ``'1.0'`` or ``'1.1'``."""
        return self._version

    @property
    def status_code(self):
        """The HTTP status code as an integer. Only for response messages."""
        return self._status_code

    @property
    def method(self):
        """The HTTP method as a string. Only for request messages."""
        return self._method

    @property
    def url(self):
        """The URL as a string. Only for request messages."""
        return self._url

    @property
    def parsed_url(self):
        """The parsed URL as a :class:`ParsedUrl` instance."""
        return self._parsed_url

    @property
    def headers(self):
        """The headers as a list of ``(name, value)`` tuples."""
        return self._headers

    delegate_method(headers, get_header)

    @property
    def charset(self):
        """The character set as parsed from the "Content-Type" header, if available."""
        return self._charset

    @property
    def body(self):
        """The message body, as a :class:`~gruvi.Stream` instance."""
        return self._body


class HttpRequest(object):
    """HTTP client request.

    Usually you do not instantiate this class directly, but use the instance
    returned by :meth:`HttpProtocol.request`. You can however start new request
    yourself by instantiating this class and passing it a protocol instance.
    """

    def __init__(self, protocol):
        self._protocol = protocol
        self._content_length = None
        self._chunked = False
        self._bytes_written = 0
        self._charset = None

    @switchpoint
    def start_request(self, method, url, headers=None, bodylen=None):
        """Start a new HTTP request.

        The optional *headers* argument contains the headers to send. It must
        be a sequence of ``(name, value)`` tuples.

        The optional *bodylen* parameter is a hint that specifies the length of
        the body that will follow. A length of -1 indicates no body, 0 means an
        empty body, and a positive number indicates the body size in bytes.
        This parameter helps determine whether to use the chunked transfer
        encoding. Normally when the body size is known chunked encoding is not used.
        """
        self._headers = headers or []
        agent = host = clen = trailer = None
        # Check the headers provided, and capture some information about the
        # request from them.
        for name, value in self._headers:
            lname = name.lower()
            # Only HTTP applications are allowed to set "hop-by-hop" headers.
            if lname in hop_by_hop:
                raise ValueError('header {} is hop-by-hop'.format(name))
            elif lname == 'user-agent':
                agent = value
            elif lname == 'host':
                host = value
            elif lname == 'content-length':
                clen = int(value)
            elif lname == 'trailer':
                trailer = parse_trailer(value)
            elif lname == 'content-type' and value.startswith('text/'):
                ctype, params = parse_content_type(value)
                self._charset = params.get('charset')
        version = self._protocol._version
        # The Host header is mandatory in 1.1. Add it if it's missing.
        if host is None and version == '1.1':
            self._headers.append(('Host', self._protocol._server_name))
        # Identify ourselves.
        if agent is None:
            self._headers.append(('User-Agent', self._protocol.identifier))
        # Check if we need to use chunked encoding due to unknown body size.
        if clen is None and bodylen is None:
            if version == '1.0':
                raise HttpError('body size unknown for HTTP/1.0')
            self._chunked = True
        self._content_length = clen
        # Check if trailers are requested and if so need to switch to chunked.
        if trailer:
            if version == '1.0':
                raise HttpError('cannot support trailers for HTTP/1.0')
            if clen is not None:
                remove_headers(self._headers, 'Content-Length')
            self._chunked = True
        self._trailer = trailer
        # Add Content-Length if we know the body size and are not using chunked.
        if not self._chunked and clen is None and bodylen >= 0:
            self._headers.append(('Content-Length', str(bodylen)))
            self._content_length = bodylen
        # Complete the "Hop by hop" headers.
        if version == '1.0':
            self._headers.append(('Connection', 'keep-alive'))
        elif version == '1.1':
            self._headers.append(('Connection', 'te'))
            self._headers.append(('TE', 'trailers'))
        if self._chunked:
            self._headers.append(('Transfer-Encoding', 'chunked'))
        # Start the request
        self._protocol._requests.append(method)
        header = create_request(version, method, url, self._headers)
        self._protocol.writer.write(header)

    @switchpoint
    def write(self, buf):
        """Write *buf* to the request body."""
        if not isinstance(buf, six.binary_type):
            raise TypeError('buf: must be a bytes instance')
        # Be careful not to write zero-length chunks as they indicate the end of a body.
        if len(buf) == 0:
            return
        if self._content_length and self._bytes_written > self._content_length:
            raise RuntimeError('wrote too many bytes ({} > {})'
                                    .format(self._bytes_written, self._content_length))
        self._bytes_written += len(buf)
        if self._chunked:
            buf = create_chunk(buf)
        self._protocol.writer.write(buf)

    @switchpoint
    def end_request(self):
        """End the request body."""
        if not self._chunked:
            return
        trailers = [(n, get_header(self._headers, n)) for n in self._trailer] \
                        if self._trailer else None
        ending = create_chunked_body_end(trailers)
        self._protocol.writer.write(ending)


class ErrorStream(object):
    """Passed to the WSGI application as environ['wsgi.errors'].

    Forwards messages to the Python logging facility.
    """

    __slots__ = ['_log']

    def __init__(self, log=None):
        self._log = log or logging.get_logger()

    def flush(self):
        pass

    def write(self, data):
        self._log.error('wsgi.errors: {}', data)

    def writelines(self, seq):
        for line in seq:
            self.write(line)


class WsgiAdapter(object):
    """WSGI Adapter"""

    def __init__(self, application):
        """
        This class adapts the WSGI callable *application* so that instances of
        this class can be used as a message handler in :class:`HttpProtocol`.
        """
        self._application = application
        self._transport = None
        self._protocol = None
        self._log = logging.get_logger()

    @switchpoint
    def send_headers(self):
        # Send out the actual headers.
        # We need to figure out the transfer encoding of the body that will
        # follow the header. Here's what we do:
        #  - If we know the body length, don't use any TE.
        #  - Otherwise, if the protocol is HTTP/1.1, use "chunked".
        #  - Otherwise, close the connection after the body is sent.
        clen = get_header(self._headers, 'Content-Length')
        version = self._message.version
        # Keep the connection alive if the request wanted it kept alive AND we
        # can keep it alive because we don't need EOF to signal end of message.
        can_chunk = version == '1.1'
        can_keep_alive = can_chunk or clen is not None or self._body_len is not None
        self._keepalive = self._message._should_keep_alive and can_keep_alive
        # The default on HTTP/1.1 is keepalive, on HTTP/1.0 it is to close.
        if version == '1.1' and not self._keepalive:
            self._headers.append(('Connection', 'close'))
        elif version == '1.0' and self._keepalive:
            self._headers.append(('Connection', 'keep-alive'))
        # Are we using chunked?
        self._chunked = can_chunk and clen is None and self._body_len is None
        if self._chunked:
            self._headers.append(('Transfer-Encoding', 'chunked'))
        elif clen is None and self._body_len is not None:
            self._headers.append(('Content-Length', str(self._body_len)))
        # Trailers..
        trailer = get_header(self._headers, 'Trailer')
        te = get_header(self._message.headers, 'TE')
        if version == '1.1' and trailer is not None and te is not None:
            tenames = [e[0].lower() for e in parse_te(te)]
            if 'trailers' in tenames:
                trailer = parse_trailer(trailer)
            else:
                remove_headers(self._headers, 'Trailer')
        self._trailer = trailer
        # Add some informational headers.
        server = get_header(self._headers, 'Server')
        if server is None:
            self._headers.append(('Server', self._protocol.identifier))
        date = get_header(self._headers, 'Date')
        if date is None:
            self._headers.append(('Date', rfc1123_date()))
        header = create_response(version, self._status, self._headers)
        self._protocol.writer.write(header)
        self._headers_sent = True

    def start_response(self, status, headers, exc_info=None):
        # Callable to be passed to the WSGI application.
        if exc_info:
            try:
                if self._headers_sent:
                    six.reraise(*exc_info)
            finally:
                exc_info = None
        elif self._status is not None:
            raise RuntimeError('response already started')
        for name, value in headers:
            if name.lower() in hop_by_hop:
                raise ValueError('header {} is hop-by-hop'.format(name))
        self._status = status
        self._headers = headers
        return self.write

    @switchpoint
    def write(self, data):
        # Callable passed to the WSGI application by start_response().
        if not isinstance(data, bytes):
            raise TypeError('data: expecting bytes instance')
        # Never write empty chunks as they signal end of body.
        if not data:
            return
        if not self._status:
            raise HttpError('WSGI handler did not call start_response()')
        if not self._headers_sent:
            self.send_headers()
        if self._chunked:
            data = create_chunk(data)
        self._protocol.writer.write(data)

    @switchpoint
    def end_response(self):
        # Finalize a response. This method must be called.
        if not self._status:
            raise HttpError('WSGI handler did not call start_response()')
        if not self._headers_sent:
            self._body_len = 0
            self.send_headers()
        if self._chunked:
            trailers = [hd for hd in self._headers if hd[0] in self._trailer] \
                            if self._trailer else None
            ending = create_chunked_body_end(trailers)
            self._protocol.writer.write(ending)
        if not self._keepalive:
            self._transport.close()

    @switchpoint
    def __call__(self, message, transport, protocol):
        # Run the WSGI handler.
        self._message = message
        if self._transport is None:
            self._transport = transport
            self._protocol = protocol
            self._sockname = transport.get_extra_info('sockname')
            self._peername = transport.get_extra_info('peername')
        self._status = None
        self._headers = None
        self._headers_sent = False
        self._chunked = False
        self._body_len = None
        self.create_environ()
        self._log.debug('request: {} {}', message.method, message.url)
        result = None
        try:
            result = self._application(self._environ, self.start_response)
            # Prevent chunking in this common case:
            if isinstance(result, list) and len(result) == 1:
                self._body_len = len(result[0])
            for chunk in result:
                self.write(chunk)
            self.end_response()
        finally:
            if hasattr(result, 'close'):
                result.close()
        ctype = get_header(self._headers, 'Content-Type', 'unknown')
        clen = get_header(self._headers, 'Content-Length', 'unknown')
        self._log.debug('response: {} ({}; {} bytes)'.format(self._status, ctype, clen))

    def create_environ(self):
        # Initialize the environment with per connection variables.
        m = self._message
        env = self._environ = {}
        # CGI variables
        env['SCRIPT_NAME'] = ''
        if isinstance(self._sockname, tuple):
            env['SERVER_NAME'] = self._protocol._server_name or self._sockname[0]
            env['SERVER_PORT'] = str(self._sockname[1])
        else:
            env['SERVER_NAME'] = self._protocol._server_name or self._sockname
            env['SERVER_PORT'] = ''
        env['SERVER_SOFTWARE'] = self._protocol.identifier
        env['SERVER_PROTOCOL'] = 'HTTP/' + m.version
        env['REQUEST_METHOD'] = m.method
        env['PATH_INFO'] = m.parsed_url[2]
        env['QUERY_STRING'] = m.parsed_url[4]
        env['REQUEST_URI'] = m.url
        for name, value in m.headers:
            if name.lower() in hop_by_hop:
                continue
            name = name.upper().replace('-', '_')
            if name not in ('CONTENT_LENGTH', 'CONTENT_TYPE'):
                name = 'HTTP_' + name
            env[name] = value
        # SSL information
        sslobj = self._transport.get_extra_info('ssl')
        cipherinfo = sslobj.cipher() if sslobj else None
        if sslobj and cipherinfo:
            env['HTTPS'] = '1'
            env['SSL_CIPHER'] = cipherinfo[0]
            env['SSL_PROTOCOL'] = cipherinfo[1]
            env['SSL_CIPHER_USEKEYSIZE'] = int(cipherinfo[2])
        # Support the de-facto X-Forwarded-For and X-Forwarded-Proto headers
        # that are added by reverse proxies.
        peername = self._transport.get_extra_info('peername')
        remote = env.get('HTTP_X_FORWARDED_FOR')
        env['REMOTE_ADDR'] = remote if remote else peername[0] \
                                        if isinstance(peername, tuple) else ''
        proto = env.get('HTTP_X_FORWARDED_PROTO')
        env['REQUEST_SCHEME'] = proto if proto else 'https' if sslobj else 'http'
        # WSGI specific variables
        env['wsgi.version'] = (1, 0)
        env['wsgi.url_scheme'] = env['REQUEST_SCHEME']
        env['wsgi.input'] = m.body
        env['wsgi.errors'] = ErrorStream(self._log)
        env['wsgi.multithread'] = True
        env['wsgi.multiprocess'] = True
        env['wsgi.run_once'] = False
        env['wsgi.charset'] = m.charset
        # Gruvi specific variables
        env['gruvi.sockname'] = self._sockname
        env['gruvi.peername'] = self._peername


class HttpProtocol(MessageProtocol):
    """HTTP protocol implementation."""

    identifier = __name__

    #: Default HTTP version.
    default_version = '1.1'

    #: Max header size. The parser keeps the header in memory during parsing.
    max_header_size = 65536

    #: Max number of body bytes to buffer. Bodies larger than this will cause
    #: the transport to be paused until the buffer is below the threshold again.
    max_buffer_size = 65536

    #: Max number of pipelined requests to keep before pausing the transport.
    max_pipeline_size = 10

    # In theory, max memory is pipeline_size * (header_size + buffer_size)

    def __init__(self, handler=None, server_side=False, server_name=None,
                 version=None, timeout=None):
        """
        The *handler* argument specifies a message handler to handle incoming
        HTTP requests. It must be a callable with the signature
        ``handler(message, transport, protocol)``.

        The *server_side* argument specifies whether this is a client or server
        side protocol. For server-side protocols, the *server_name* argument
        can be used to provide a server name. If not provided, then the socket
        name of the listening socket will be used.
        """
        super(HttpProtocol, self).__init__(handler, timeout=timeout)
        if server_side and handler is None:
            raise ValueError('need a handler for server side protocol')
        self._handler = handler
        self._server_side = server_side
        self._server_name = server_name
        self._version = self.default_version if version is None else version
        if self._version not in ('1.0', '1.1'):
            raise ValueError('unsupported HTTP version: {!r}'.format(version))
        self._timeout = timeout
        self._create_parser()
        self._requests = []
        self._response = None
        self._writer = None
        self._message = None
        self._error = None

    def _create_parser(self):
        # Create a new CFFI http and parser.
        self._parser = ffi.new('http_parser *')
        self._cdata = ffi.new_handle(self)  # store in instance to keep alive:
        self._parser.data = self._cdata     # struct field doesn't take reference
        kind = lib.HTTP_REQUEST if self._server_side else lib.HTTP_RESPONSE
        lib.http_parser_init(self._parser, kind)
        self._urlparser = ffi.new('struct http_parser_url *')

    def _append_header_name(self, buf):
        # Add a chunk to a header name.
        if self._header_value:
            # This starts a new header: stash away the previous one.
            # The header might be part of the http headers or trailers.
            header = (ba2s(self._header_name), ba2s(self._header_value))
            self._message.headers.append(header)
            # Try to capture the charset for text bodies.
            if header[0].lower() == 'content-type':
                ctype, params = parse_content_type(header[1])
                if ctype.startswith('text/'):
                    self._message._charset = params.get('charset')
            del self._header_name[:]
            del self._header_value[:]
        self._header_name.extend(buf)

    def _append_header_value(self, buf):
        # Add a chunk to a header value.
        self._header_value.extend(buf)

    # Parser callbacks. These are Python methods called by http-parser C code
    # via CFFI. Callbacks are run in the hub fiber, and we only do parsing
    # here, no handlers are run. For server protocols we stash away the result
    # in a queue to be processed in a dispatcher fiber (one per protocol).
    #
    # Callbacks return 0 for success, 1 for error.
    #
    # Also note that these are static methods that get a reference to their
    # instance via the parser state (parser.data).

    @ffi.callback('http_cb')
    def on_message_begin(parser):
        # http-parser callback: prepare for a new message
        self = ffi.from_handle(parser.data)
        self._url = bytearray()
        self._header = bytearray()
        self._header_name = bytearray()
        self._header_value = bytearray()
        self._header_size = 0
        self._message = HttpMessage()
        lib.http_parser_url_init(self._urlparser)
        return 0

    @ffi.callback('http_data_cb')
    def on_url(parser, at, length):
        # http-parser callback: got a piece of the URL
        self = ffi.from_handle(parser.data)
        self._header_size += length
        if self._header_size > self.max_header_size:
            self._error = HttpError('HTTP header too large')
            return 1
        self._url.extend(ffi.buffer(at, length))
        return 0

    @ffi.callback('http_data_cb')
    def on_header_name(parser, at, length):
        # http-parser callback: got a piece of a header name
        self = ffi.from_handle(parser.data)
        self._header_size += length
        if self._header_size > self.max_header_size:
            self._error = HttpError('HTTP header too large')
            return 1
        buf = ffi.buffer(at, length)
        self._append_header_name(buf)
        return 0

    @ffi.callback('http_data_cb')
    def on_header_value(parser, at, length):
        # http-parser callback: got a piece of a header value
        self = ffi.from_handle(parser.data)
        self._header_size += length
        if self._header_size > self.max_header_size:
            self._error = HttpError('HTTP header too large')
            return 1
        buf = ffi.buffer(at, length)
        self._append_header_value(buf)
        return 0

    @ffi.callback('http_cb')
    def on_headers_complete(parser):
        # http-parser callback: the HTTP header is complete. This is the point
        # where we hand off the message to our consumer. Going forward,
        # on_body() will continue to write chunks of the body to message.body.
        self = ffi.from_handle(parser.data)
        self._append_header_name(b'')
        m = self._message
        m._message_type = lib.http_message_type(parser)
        m._version = '{}.{}'.format(parser.http_major, parser.http_minor)
        if self._server_side:
            m._method = _http_methods.get(lib.http_method(parser), '<unknown>')
            m._url = ba2s(self._url)
            res = lib.http_parser_parse_url(ffi.from_buffer(self._url), len(self._url),
                                            m._method == 'CONNECT', self._urlparser)
            assert res == 0   # URL was already validated by http-parser
            m._parsed_url = ParsedUrl.from_parser(self._urlparser, m._url)
        else:
            m._status_code = lib.http_status_code(parser)
        m._should_keep_alive = lib.http_should_keep_alive(parser)
        m._body = Stream(self._transport, 'r')
        m._body.buffer.set_buffer_limits(self.max_buffer_size)
        # Make the message available on the queue.
        self._queue.put_nowait(m)
        # Return 1 if this is a response to a HEAD request. This is a hint to
        # the parser that no body will follow. Normally the parser deduce from
        # the headers whether body will follow (either Content-Length or
        # Transfer-Encoding is present). But not so with HEAD, as the response
        # is exactly identical to GET but without the body.
        if not self._server_side and self._requests and self._requests.pop(0) == 'HEAD':
            return 1
        return 0

    @ffi.callback('http_data_cb')
    def on_body(parser, at, length):
        # http-parser callback: got a body chunk
        self = ffi.from_handle(parser.data)
        # StreamBuffer.feed() may pause the transport here if the buffer size is exceeded.
        self._message.body.buffer.feed(bytes(ffi.buffer(at, length)))
        return 0

    @ffi.callback('http_cb')
    def on_message_complete(parser):
        # http-parser callback: the http request or response ended
        # complete any trailers that might be present
        self = ffi.from_handle(parser.data)
        self._append_header_name(b'')
        self._message.body.buffer.feed_eof()
        self._maybe_pause_transport()
        return 0

    # The settings object is shared between all protocol instances.

    _settings = ffi.new('http_parser_settings *')
    _settings.on_message_begin = on_message_begin
    _settings.on_url = on_url
    _settings.on_header_field = on_header_name
    _settings.on_header_value = on_header_value
    _settings.on_headers_complete = on_headers_complete
    _settings.on_body = on_body
    _settings.on_message_complete = on_message_complete

    def connection_made(self, transport):
        # Protocol callback
        super(HttpProtocol, self).connection_made(transport)
        self._transport = transport
        self._writer = Stream(transport, 'w')

    def data_received(self, data):
        # Protocol callback
        nbytes = lib.http_parser_execute(self._parser, self._settings, data, len(data))
        if nbytes != len(data):
            msg = cd2s(lib.http_errno_name(lib.http_errno(self._parser)))
            self._log.debug('http_parser_execute(): {}'.format(msg))
            self._error = HttpError('parse error: {}'.format(msg))
            if self._message:
                self._message.body.buffer.feed_error(self._error)
            self._transport.close()

    def connection_lost(self, exc):
        # Protocol callback
        # Feed the EOF to the parser. It will tell us it if was unexpected.
        super(HttpProtocol, self).connection_lost(exc)
        nbytes = lib.http_parser_execute(self._parser, self._settings, b'', 0)
        if nbytes != 0:
            msg = cd2s(lib.http_errno_name(lib.http_errno(self._parser)))
            self._log.debug('http_parser_execute(): {}'.format(msg))
            if exc is None:
                exc = HttpError('parse error: {}'.format(msg))
            if self._message:
                self._message.body.buffer.feed_error(exc)
        self._error = exc

    @property
    def writer(self):
        """A :class:`~gruvi.Stream` instance for writing directly to the
        underlying transport."""
        return self._writer

    @switchpoint
    def request(self, method, url, headers=None, body=None):
        """Make a new HTTP request.

        The *method* argument is the HTTP method as a string, for example
        ``'GET'`` or ``'POST'``. The *url* argument specifies the URL.

        The optional *headers* argument specifies extra HTTP headers to use in
        the request. It must be a sequence of ``(name, value)`` tuples.

        The optional *body* argument may be used to include a body in the
        request. It must be a ``bytes`` instance, a file-like object opened in
        binary mode, or an iterable producing ``bytes`` instances.  To send
        potentially large bodies, use the file or iterator interfaces. This has
        the benefit that only a single chunk is kept in memory at a time.

        The response to the request can be obtained by calling the
        :meth:`getresponse` method. You may make multiple requests before
        reading a response. For every request that you make however, you must
        call :meth:`getresponse` exactly once. The remote HTTP implementation
        will send by the responses in the same order as the requests.

        This method will use the "chunked" transfer encoding if here is a body
        and the body size is unknown ahead of time. This happens when the file
        or interator interface is used in the abence of a "Content-Length"
        header.
        """
        if self._error:
            raise compat.saved_exc(self._error)
        elif self._transport is None:
            raise HttpError('not connected')
        request = HttpRequest(self)
        bodylen = -1 if body is None else \
                        len(body) if isinstance(body, bytes) else None
        request.start_request(method, url, headers, bodylen)
        if isinstance(body, bytes):
            request.write(body)
        elif hasattr(body, 'read'):
            while True:
                chunk = body.read(4096)
                if not chunk:
                    break
                request.write(chunk)
        elif hasattr(body, '__iter__'):
            for chunk in body:
                request.write(chunk)
        request.end_request()

    @switchpoint
    def getresponse(self):
        """Wait for and return a HTTP response.

        The return value will be a :class:`HttpMessage`. When this method
        returns only the response header has been read. The response body can
        be read using :meth:`~gruvi.Stream.read` and similar methods on
        the message :attr:`~HttpMessage.body`.

        Note that if you use persistent connections (the default), it is
        required that you read the entire body of each response. If you don't
        then deadlocks may occur.
        """
        if self._error:
            raise compat.saved_exc(self._error)
        elif self._transport is None:
            raise HttpError('not connected')
        message = self._queue.get(timeout=self._timeout)
        if isinstance(message, Exception):
            raise compat.saved_exc(message)
        return message


class HttpClient(Client):
    """HTTP client."""

    def __init__(self, version=None, timeout=None):
        """
        The optional *version* argument specifies the HTTP version to use. The
        default is :attr:`HttpProtocol.default_version`.

        The optional *timeout* argument specifies the timeout for various
        network and protocol operations.
        """
        protocol_factory = functools.partial(HttpProtocol, version=version)
        super(HttpClient, self).__init__(protocol_factory, timeout=timeout)
        self._server_name = None

    @docfrom(Client.connect)
    def connect(self, address, **kwargs):
        # Capture the host name that we are connecting to. We need this for
        # generating "Host" headers in HTTP/1.1
        if self._server_name is None and isinstance(address, tuple):
            host, port = address[:2]  # len(address) == 4 for IPv6
            if port != default_ports['https' if 'ssl' in kwargs else 'http']:
                host = '{}:{}'.format(host, port)
            self._server_name = host
        super(HttpClient, self).connect(address, **kwargs)
        self._protocol._server_name = self._server_name

    protocol = Client.protocol

    delegate_method(protocol, HttpProtocol.request)
    delegate_method(protocol, HttpProtocol.getresponse)


class HttpServer(Server):
    """HTTP server."""

    #: The default adapter to use.
    default_adapter = WsgiAdapter

    def __init__(self, application, server_name=None, adapter=None, timeout=None):
        """The *application* argument is the web application to expose on this
        server. The application is wrapped in *adapter* to create a message
        handler as required by :class:`HttpProtocol`. By default the adapter in
        :attr:`default_adapter` is used.

        The optional *server_name* argument specifies the server name. The
        default is to use the host portion of the address passed to
        :meth:`~gruvi.Server.listen`. The server name is made available to WSGI
        applications as the $SERVER_NAME environment variable.

        The optional *timeout* argument specifies the timeout for various
        network and protocol operations.
        """
        adapter = self.default_adapter if adapter is None else adapter
        def handler(*args):
            return adapter(application)(*args)
        protocol_factory = functools.partial(HttpProtocol, handler,
                                             server_side=True, server_name=server_name)
        super(HttpServer, self).__init__(protocol_factory, timeout)
        self._server_name = server_name

    def connection_made(self, transport, protocol):
        if protocol._server_name is None:
            protocol._server_name = self._server_name

    def listen(self, address, **kwargs):
        # Capture the first listen address to provide for a default server name.
        if not self._server_name and isinstance(address, tuple):
            host, port = address[:2]  # len(address) == 4 for IPv6
            if port != default_ports['https' if 'ssl' in kwargs else 'http']:
                host = '{}:{}'.format(host, port)
            self._server_name = host
        super(HttpServer, self).listen(address, **kwargs)



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

from . import logging

# Many objects in Gruvi need to keep a list of callbacks. This module provides
# a few utility functions to do this in an efficient way.
#
# Callbacks are stored in a linked list. This allows to iterate over them in
# insertion order and also allows removal from the middle assuming the node
# handle is kept by the caller.

# Linked list implementation:

if __debug__:

    def fmtnode(node):
        return '<Node(prev={:#x}, next={:#x}, value={!r})>' \
                    .format(id(node._prev), id(node._next), node.value)

    def dump(dll):
        print('== Dumping dllist {!r}'.format(dll))
        print('Size: {}'.format(dll._size))
        print('First: {}'.format(fmtnode(dll.first) if dll.first else 'None'))
        print('Last: {}'.format(fmtnode(dll.last) if dll.last else 'None'))
        print('Nodes:')
        count = 0
        node = dll.first
        while node is not None:
            print('- {} [{}]'.format(fmtnode(node), count))
            node = node._next
            count += 1
        print('Total nodes: {}'.format(count))

    def check(dll):
        if dll.first is None:
            assert dll.last is None
            assert dll._size == 0
            return
        node = dll.first
        assert node._list is dll
        assert node._prev is None
        nnode = node._next
        count = 1
        while nnode is not None:
            assert nnode._list is dll
            assert nnode._prev is node
            node, nnode = nnode, nnode._next
            count += 1
        assert node is dll.last
        assert count == dll._size


class Node(object):
    """A node in a doubly linked list."""

    __slots__ = ('_prev', '_next', '_list', 'callback', 'args')

    def __init__(self, callback=None, args=None):
        self._prev = None
        self._next = None
        self._list = None
        self.callback = callback
        self.args = args


class dllist(object):
    """A doubly linked list."""

    __slots__ = ('_first', '_last', '_size')

    def __init__(self):
        self._first = None
        self._last = None
        self._size = 0

    @property
    def first(self):
        """The first node in the list."""
        return self._first

    @property
    def last(self):
        """The last node in the list."""
        return self._last

    def __len__(self):
        return self._size

    def __contains__(self, node):
        """Return whether *node* is contained in the list."""
        if not isinstance(node, Node):
            raise TypeError('expecting Node instance')
        return node._list is self

    def remove(self, node):
        """Remove a node from the list."""
        if not isinstance(node, Node):
            raise TypeError('expecting Node instance')
        if node._list is None:
            return
        if node._list is not self:
            raise RuntimeError('node is not contained in list')
        if node._next is None:
            self._last = node._prev  # last node
        else:
            node._next._prev = node._prev
        if node._prev is None:
            self._first = node._next  # first node
        else:
            node._prev._next = node._next
        node._list = node._prev = node._next = None
        self._size -= 1

    def insert(self, node, before=None):
        """Insert a new node in the list.

        If *before* is specified, the new node is inserted before this node.
        Otherwise, the node is inserted at the end of the list.
        """
        node._list = self
        if self._first is None:
            self._first = self._last = node  # first node in list
            self._size += 1
            return node
        if before is None:
            self._last._next = node  # insert as last node
            node._prev = self._last
            self._last = node
        else:
            node._next = before
            node._prev = before._prev
            if node._prev:
                node._prev._next = node
            else:
                self._first = node  # inserting as first node
            node._next._prev = node
        self._size += 1
        return node

    def __iter__(self):
        """Return an iterator/generator that yields all nodes.

        Note: it is safe to remove the current node while iterating but you
        should not remove the next one.
        """
        node = self._first
        while node is not None:
            next_node = node._next
            yield node
            node = next_node

    def clear(self):
        """Remove all nodes from the list."""
        node = self._first
        while node is not None:
            next_node = node._next
            node._list = node._prev = node._next = None
            node = next_node
        self._size = 0


# Callback utilities. These utilities add some more optimizations.
#
# * The add/remove callback functions are implemented as functions operating on
#   an object. They store the callbacks in the '_callbacks' property of the
#   object.
# * A single callback is stored directly in the object as a Node.
# * Multiple calbacks are stored as a dllist of nodes.


def add_callback(obj, callback, args=()):
    """Add a callback to an object."""
    callbacks = obj._callbacks
    node = Node(callback, args)
    # Store a single callback directly in _callbacks
    if callbacks is None:
        obj._callbacks = node
        return node
    # Otherwise use a dllist.
    if not isinstance(callbacks, dllist):
        obj._callbacks = dllist()
        obj._callbacks.insert(callbacks)
        callbacks = obj._callbacks
    callbacks.insert(node)
    return node


def remove_callback(obj, handle):
    """Remove a callback from an object."""
    callbacks = obj._callbacks
    if callbacks is handle:
        obj._callbacks = None
    elif isinstance(callbacks, dllist):
        callbacks.remove(handle)
        if not callbacks:
            obj._callbacks = None


def has_callback(obj, handle):
    """Return whether a callback is currently registered for an object."""
    callbacks = obj._callbacks
    if not callbacks:
        return False
    if isinstance(callbacks, Node):
        return handle is callbacks
    else:
        return handle in callbacks


def pop_callback(obj):
    """Pop a single callback."""
    callbacks = obj._callbacks
    if not callbacks:
        return
    if isinstance(callbacks, Node):
        node = callbacks
        obj._callbacks = None
    else:
        node = callbacks.first
        callbacks.remove(node)
        if not callbacks:
            obj._callbacks = None
    return node.callback, node.args


def clear_callbacks(obj):
    """Remove all callbacks from an object."""
    callbacks = obj._callbacks
    if isinstance(callbacks, dllist):
        # Help the garbage collector by clearing all links.
        callbacks.clear()
    obj._callbacks = None


def walk_callbacks(obj, func, log=None):
    """Call func(callback, args) for all callbacks and keep only those
    callbacks for which the function returns True."""
    callbacks = obj._callbacks
    if isinstance(callbacks, Node):
        node = callbacks
        try:
            if not func(node.callback, node.args):
                obj._callbacks = None
        except Exception:
            if log is None:
                log = logging.get_logger()
            log.exception('uncaught exception in callback')
    elif isinstance(callbacks, dllist):
        for node in callbacks:
            try:
                if func(node.callback, node.args):
                    continue
                callbacks.remove(node)
            except Exception:
                if log is None:
                    log = logging.get_logger()
                log.exception('uncaught exception in callback')
        if not callbacks:
            obj._callbacks = None


def run_callbacks(obj, log=None):
    """Run callbacks."""
    def run_callback(callback, args):
        return callback(*args)
    return walk_callbacks(obj, run_callback, log)



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import os
import six
import socket
import unittest
from unittest import SkipTest

import gruvi
from gruvi.vendor import txdbus
from gruvi.dbus import DbusError, DbusMethodCallError
from gruvi.dbus import DbusProtocol, DbusClient, DbusServer
from gruvi.dbus import parse_dbus_header, TxdbusAuthenticator
from gruvi.transports import TransportError

from support import UnitTest, MockTransport


class TestParseDbusHeader(UnitTest):

    def test_simple(self):
        m = b'l\1\0\1\0\0\0\0\1\0\0\0\0\0\0\0'
        self.assertEqual(parse_dbus_header(m), len(m))

    def test_big_endian(self):
        m = b'B\1\0\1\0\0\0\0\0\0\0\1\0\0\0\0'
        self.assertEqual(parse_dbus_header(m), len(m))

    def test_header_array(self):
        m = b'l\1\0\1\0\0\0\0\1\0\0\0\10\0\0\0h2345678'
        self.assertEqual(parse_dbus_header(m), len(m))
        for l in range(16, len(m)):
            self.assertEqual(parse_dbus_header(m[:l]), len(m))

    def test_padding(self):
        m = b'l\1\0\1\0\0\0\0\1\0\0\0\11\0\0\0h234567812345678'
        self.assertEqual(parse_dbus_header(m), len(m))
        for l in range(16, len(m)):
            self.assertEqual(parse_dbus_header(m[:l]), len(m))

    def test_body_size(self):
        m = b'l\1\0\1\4\0\0\0\1\0\0\0\0\0\0\0b234'
        self.assertEqual(parse_dbus_header(m), len(m))
        for l in range(16, len(m)):
            self.assertEqual(parse_dbus_header(m[:l]), len(m))

    def test_illegal_endian(self):
        m = b'L\1\0\1\0\0\0\0\1\0\0\0\0\0\0\0'
        self.assertRaises(ValueError, parse_dbus_header, m)
        m = b'b\1\0\1\0\0\0\0\1\0\0\0\0\0\0\0'
        self.assertRaises(ValueError, parse_dbus_header, m)

    def test_illegal_type(self):
        m = b'l\0\0\1\0\0\0\0\1\0\0\0\0\0\0\0'
        self.assertRaises(ValueError, parse_dbus_header, m)
        m = b'l\5\0\1\0\0\0\0\1\0\0\0\0\0\0\0'
        self.assertRaises(ValueError, parse_dbus_header, m)

    def test_illegal_serial(self):
        m = b'l\1\0\1\0\0\0\0\0\0\0\0\0\0\0\0'
        self.assertRaises(ValueError, parse_dbus_header, m)


class TestDbusProtocol(UnitTest):

    def setUp(self):
        super(TestDbusProtocol, self).setUp()
        self.messages = []
        self.protocols = []

    def store_messages(self, message, transport, protocol):
        self.messages.append(message)
        self.protocols.append(protocol)

    def store_and_echo_messages(self, message, transport, protocol):
        self.messages.append(message)
        self.protocols.append(protocol)
        response = txdbus.SignalMessage(message.path, message.member, message.interface,
                                        signature=message.signature, body=message.body)
        protocol.send_message(response)

    def test_auth_missing_creds_byte(self):
        # The first thing a client should send to the server is a '\0' byte. If
        # not, the server should close the connection.
        transport = MockTransport()
        protocol = DbusProtocol(server_side=True)
        transport.start(protocol)
        self.assertFalse(transport._closed.is_set())
        protocol.data_received(b'\1')
        self.assertIsInstance(protocol._error, DbusError)
        self.assertTrue(transport._closed.is_set())

    def test_auth_non_ascii(self):
        # After the '\0' byte, an authentiction phase happens. The
        # authentication protocol is line based and all lines should be ascii.
        transport = MockTransport()
        protocol = DbusProtocol(server_side=True)
        transport.start(protocol)
        self.assertFalse(transport._closed.is_set())
        protocol.data_received(b'\0\xff\r\n')
        self.assertIsInstance(protocol._error, DbusError)
        self.assertTrue(transport._closed.is_set())

    def test_auth_long_line(self):
        # An authentication line should not exceed the maximum line size.
        transport = MockTransport()
        protocol = DbusProtocol(server_side=True, server_guid='foo')
        protocol.max_line_size = 5
        transport.start(protocol)
        self.assertFalse(transport._closed.is_set())
        protocol.data_received(b'\0AUTH ANONYMOUS\r\n')
        self.assertIsInstance(protocol._error, DbusError)
        self.assertTrue(transport._closed.is_set())

    def test_auth_ok(self):
        # Test anonymous authenication. Ensure that the server GUID is
        # correctly sent back.
        transport = MockTransport()
        protocol = DbusProtocol(server_side=True, server_guid='foo')
        transport.start(protocol)
        protocol.data_received(b'\0AUTH ANONYMOUS\r\nBEGIN\r\n')
        buf = transport.buffer.getvalue()
        self.assertTrue(buf.startswith(b'OK foo'))
        auth = protocol._authenticator
        self.assertTrue(auth.authenticationSucceeded())
        self.assertTrue(auth.getGUID(), 'foo')
        self.assertFalse(transport._closed.is_set())

    def test_missing_hello(self):
        # After authentication, the first message should be a "Hello".
        # Otherwise, the server should close the connection.
        transport = MockTransport()
        protocol = DbusProtocol(self.store_messages, server_side=True, server_guid='foo')
        transport.start(protocol)
        protocol.data_received(b'\0AUTH ANONYMOUS\r\nBEGIN\r\n')
        message = txdbus.MethodCallMessage('/my/path', 'Method')
        auth = protocol._authenticator
        self.assertTrue(auth.authenticationSucceeded())
        protocol.data_received(message.rawMessage)
        self.assertIsInstance(protocol._error, DbusError)
        self.assertTrue(transport._closed.is_set())

    def test_send_message(self):
        # After the "Hello" message, it should be possible to send other
        # messages.
        transport = MockTransport()
        protocol = DbusProtocol(self.store_messages, server_side=True, server_guid='foo')
        transport.start(protocol)
        protocol.data_received(b'\0AUTH ANONYMOUS\r\nBEGIN\r\n')
        auth = protocol._authenticator
        self.assertTrue(auth.authenticationSucceeded())
        message = txdbus.MethodCallMessage('/org/freedesktop/DBus', 'Hello',
                        interface='org.freedesktop.DBus', destination='org.freedesktop.DBus')
        protocol.data_received(message.rawMessage)
        gruvi.sleep(0)
        self.assertIsNone(protocol._error)
        self.assertFalse(transport._closed.is_set())
        self.assertTrue(protocol._name_acquired)
        self.assertEqual(len(self.messages), 0)
        message = txdbus.MethodCallMessage('/my/path', 'Method')
        protocol.data_received(message.rawMessage)
        gruvi.sleep(0)
        self.assertIsNone(protocol._error)
        self.assertFalse(transport._closed.is_set())
        self.assertEqual(len(self.messages), 1)
        self.assertEqual(self.messages[0].path, '/my/path')
        self.assertEqual(self.messages[0].member, 'Method')
        self.assertEqual(self.protocols, [protocol])

    def test_send_message_incremental(self):
        # Send a message byte by byte. The protocol should be able process it.
        transport = MockTransport()
        protocol = DbusProtocol(self.store_messages, server_side=True, server_guid='foo')
        transport.start(protocol)
        authexchange = b'\0AUTH ANONYMOUS\r\nBEGIN\r\n'
        for i in range(len(authexchange)):
            protocol.data_received(authexchange[i:i+1])
        auth = protocol._authenticator
        self.assertTrue(auth.authenticationSucceeded())
        message = txdbus.MethodCallMessage('/org/freedesktop/DBus', 'Hello',
                        interface='org.freedesktop.DBus', destination='org.freedesktop.DBus')
        for i in range(len(message.rawMessage)):
            protocol.data_received(message.rawMessage[i:i+1])
        gruvi.sleep(0)
        self.assertIsNone(protocol._error)
        self.assertFalse(transport._closed.is_set())
        self.assertEqual(len(self.messages), 0)
        message = txdbus.MethodCallMessage('/my/path', 'Method')
        for i in range(len(message.rawMessage)):
            protocol.data_received(message.rawMessage[i:i+1])
        gruvi.sleep(0)
        self.assertIsNone(protocol._error)
        self.assertFalse(transport._closed.is_set())
        self.assertEqual(len(self.messages), 1)
        self.assertEqual(self.messages[0].path, '/my/path')
        self.assertEqual(self.messages[0].member, 'Method')
        self.assertEqual(self.protocols, [protocol])

    def test_send_message_too_large(self):
        # Send a message that exceeds the maximum message size. The connection
        # should be closed.
        transport = MockTransport()
        protocol = DbusProtocol(self.store_messages, server_side=True, server_guid='foo')
        transport.start(protocol)
        protocol.data_received(b'\0AUTH ANONYMOUS\r\nBEGIN\r\n')
        message = txdbus.MethodCallMessage('/org/freedesktop/DBus', 'Hello',
                        interface='org.freedesktop.DBus', destination='org.freedesktop.DBus')
        protocol.data_received(message.rawMessage)
        gruvi.sleep(0)
        self.assertTrue(protocol._name_acquired)
        # Send a signal with a size equal to the high-water mark. This should work.
        message = txdbus.SignalMessage('/my/path', 'Signal', 'my.iface',
                                       signature='s', body=['x'*100])
        msglen = len(message.rawMessage)
        self.assertGreater(msglen, 100)
        protocol.max_message_size = msglen
        protocol.data_received(message.rawMessage)
        gruvi.sleep(0)
        self.assertIsNone(protocol._error)
        self.assertFalse(transport._closed.is_set())
        self.assertEqual(len(self.messages), 1)
        # Now send a signal with a size larger than the high-water mark. This should fail.
        message = txdbus.SignalMessage('/my/path', 'Signal', 'my.iface',
                                       signature='s', body=['x'*100])
        msglen = len(message.rawMessage)
        protocol.max_message_size = msglen-1
        protocol.data_received(message.rawMessage)
        gruvi.sleep(0)
        self.assertIsInstance(protocol._error, DbusError)
        self.assertTrue(transport._closed.is_set())
        self.assertEqual(len(self.messages), 1)

    def test_read_write_flow_control(self):
        # Send a lot of messages filling up the protocol read buffer.
        transport = MockTransport()
        protocol = DbusProtocol(self.store_and_echo_messages, server_side=True)
        transport.start(protocol)
        protocol.data_received(b'\0AUTH ANONYMOUS\r\nBEGIN\r\n')
        auth = protocol._authenticator
        self.assertTrue(auth.authenticationSucceeded())
        message = txdbus.MethodCallMessage('/org/freedesktop/DBus', 'Hello',
                        interface='org.freedesktop.DBus', destination='org.freedesktop.DBus')
        protocol.data_received(message.rawMessage)
        gruvi.sleep(0)
        self.assertTrue(protocol._name_acquired.is_set())
        interrupted = 0
        message = txdbus.SignalMessage('/my/path', 'Signal', 'my.iface',
                                       signature='s', body=['x'*100])
        msglen = len(message.rawMessage)
        protocol.max_queue_size = 10
        transport.drain()
        transport.set_write_buffer_limits(7*msglen)
        for i in range(100):
            # Fill up protocol message queue
            protocol.data_received(message.rawMessage)
            if transport._reading:
                continue
            interrupted += 1
            self.assertEqual(protocol._queue.qsize(), 10)
            # Run the dispatcher to fill up the transport write buffer
            gruvi.sleep(0)
            # Now the write buffer is full and the read buffer still contains
            # some entries because it is larger.
            self.assertGreater(protocol._queue.qsize(), 0)
            self.assertFalse(transport._can_write.is_set())
            transport.drain()
        # Should be interrupted > 10 times. The write buffer is the limiting factor
        # not the read buffer.
        self.assertGreater(interrupted, 10)


def echo_app(message, transport, protocol):
    # Test application that echos D-Bus arguments
    if not isinstance(message, txdbus.MethodCallMessage):
        return
    if message.member == 'Echo':
        reply = txdbus.MethodReturnMessage(message.serial, signature=message.signature,
                                           body=message.body)
    elif message.member == 'Error':
        reply = txdbus.ErrorMessage('Echo.Error', message.serial, signature=message.signature,
                                    body=message.body)
    else:
        return
    protocol.send_message(reply)


class TestGruviDbus(UnitTest):

    def setUp(self):
        super(TestGruviDbus, self).setUp()
        TxdbusAuthenticator.cookie_dir = self.tempdir

    def test_auth_pipe(self):
        # Test that authentication works over a Pipe.
        server = DbusServer(echo_app)
        addr = 'unix:path=' + self.pipename()
        server.listen(addr)
        client = DbusClient()
        client.connect(addr)
        cproto = client.protocol
        cauth = cproto._authenticator
        sproto = list(server.connections)[0][1]
        sauth = sproto._authenticator
        self.assertTrue(cauth.authenticationSucceeded())
        self.assertTrue(sauth.authenticationSucceeded())
        self.assertIsInstance(cproto.server_guid, six.text_type)
        self.assertTrue(cproto.server_guid.isalnum())
        self.assertEqual(cproto.server_guid, cauth.getGUID())
        self.assertEqual(cproto.server_guid, sproto.server_guid)
        self.assertEqual(sproto.server_guid, sauth.getGUID())
        self.assertEqual(cauth.getMechanismName(), sauth.getMechanismName())
        if hasattr(socket, 'SO_PEERCRED'):
            self.assertEqual(cauth.getMechanismName(), 'EXTERNAL')
        elif hasattr(os, 'fork'):
            self.assertEqual(cauth.getMechanismName(), 'DBUS_COOKIE_SHA1')
        else:
            self.assertEqual(cauth.getMechanismName(), 'ANONYMOUS')
        client.close()
        server.close()

    def test_auth_tcp(self):
        # Test that authentication works over TCP
        server = DbusServer(echo_app)
        addr = 'tcp:host=127.0.0.1,port=0'
        server.listen(addr)
        client = DbusClient()
        client.connect(server.addresses[0])
        cproto = client.protocol
        cauth = cproto._authenticator
        sproto = list(server.connections)[0][1]
        sauth = sproto._authenticator
        self.assertTrue(cauth.authenticationSucceeded())
        self.assertTrue(sauth.authenticationSucceeded())
        self.assertIsInstance(cproto.server_guid, six.text_type)
        self.assertTrue(cproto.server_guid.isalnum())
        self.assertEqual(cproto.server_guid, cauth.getGUID())
        self.assertEqual(cproto.server_guid, sproto.server_guid)
        self.assertEqual(sproto.server_guid, sauth.getGUID())
        self.assertEqual(cauth.getMechanismName(), sauth.getMechanismName())
        if hasattr(os, 'fork'):
            self.assertEqual(cauth.getMechanismName(), 'DBUS_COOKIE_SHA1')
        else:
            self.assertEqual(cauth.getMechanismName(), 'ANONYMOUS')
        client.close()
        server.close()

    def test_get_unique_name(self):
        # Ensure that get_unique_name() works client and server side
        server = DbusServer(echo_app)
        addr = 'unix:path=' + self.pipename()
        server.listen(addr)
        client = DbusClient()
        client.connect(addr)
        unique_name = client.get_unique_name()
        self.assertIsInstance(unique_name, six.text_type)
        self.assertTrue(unique_name.startswith(':'))
        sproto = list(server.connections)[0][1]
        self.assertEqual(unique_name, sproto.get_unique_name())
        server.close()
        client.close()

    def test_call_method(self):
        # Ensure that calling a method over a Unix socket works.
        server = DbusServer(echo_app)
        addr = 'unix:path=' + self.pipename()
        server.listen(addr)
        client = DbusClient()
        client.connect(addr)
        result = client.call_method('bus.name', '/path', 'my.iface', 'Echo')
        self.assertEqual(result, ())
        server.close()
        client.close()

    def test_call_method_tcp(self):
        # Ensure that calling a method over TCP works.
        server = DbusServer(echo_app)
        addr = 'tcp:host=127.0.0.1,port=0'
        server.listen(addr)
        client = DbusClient()
        client.connect(server.addresses[0])
        result = client.call_method('bus.name', '/path', 'my.iface', 'Echo')
        self.assertEqual(result, ())
        server.close()
        client.close()

    def test_call_method_str_args(self):
        # Ensure that calling a method with string arguments works.
        server = DbusServer(echo_app)
        addr = 'unix:path=' + self.pipename()
        server.listen(addr)
        client = DbusClient()
        client.connect(addr)
        result = client.call_method('bus.name', '/path', 'my.iface', 'Echo',
                                    signature='s', args=['foo'])
        self.assertEqual(result, ('foo',))
        result = client.call_method('bus.name', '/path', 'my.iface', 'Echo',
                                    signature='ss', args=['foo', 'bar'])
        self.assertEqual(result, ('foo', 'bar'))
        server.close()
        client.close()

    def test_call_method_int_args(self):
        # Ensure that calling a method with integer arguments works.
        server = DbusServer(echo_app)
        addr = 'unix:path=' + self.pipename()
        server.listen(addr)
        client = DbusClient()
        client.connect(addr)
        result = client.call_method('bus.name', '/path', 'my.iface', 'Echo',
                                    signature='i', args=[1])
        self.assertEqual(result, (1,))
        result = client.call_method('bus.name', '/path', 'my.iface', 'Echo',
                                    signature='ii', args=[1, 2])
        self.assertEqual(result, (1, 2))
        server.close()
        client.close()

    def test_call_method_error(self):
        # Ensure that a method can return an error and that in this case a
        # DbusMethodCallError is raised.
        server = DbusServer(echo_app)
        addr = 'unix:path=' + self.pipename()
        server.listen(addr)
        client = DbusClient()
        client.connect(addr)
        exc = self.assertRaises(DbusMethodCallError, client.call_method,
                               'bus.name', '/path', 'my.iface', 'Error')
        self.assertEqual(exc.error, 'Echo.Error')
        self.assertEqual(exc.args, ())
        server.close()
        client.close()

    def test_call_method_error_args(self):
        # Call a method that will return an error with arguments. The arguments
        # should be available from the exception.
        server = DbusServer(echo_app)
        addr = 'unix:path=' + self.pipename()
        server.listen(addr)
        client = DbusClient()
        client.connect(addr)
        exc = self.assertRaises(DbusMethodCallError, client.call_method,
                               'bus.name', '/path', 'my.iface', 'Error',
                               signature='ss', args=('foo', 'bar'))
        self.assertEqual(exc.error, 'Echo.Error')
        self.assertEqual(exc.args, ('foo', 'bar'))
        server.close()
        client.close()

    def test_send_garbage(self):
        # Send random garbage and ensure the connection gets dropped.
        server = DbusServer(echo_app)
        addr = 'unix:path=' + self.pipename()
        server.listen(addr)
        client = DbusClient()
        client.connect(addr)
        exc = None
        try:
            while True:
                chunk = os.urandom(100)
                client.transport.write(chunk)
                gruvi.sleep(0)
        except Exception as e:
            exc = e
        self.assertIsInstance(exc, TransportError)
        server.close()
        client.close()

    def test_connection_limit(self):
        # Establish more connections than the DBUS server is willing to accept.
        # The connections should be closed.
        server = DbusServer(echo_app)
        addr = 'unix:path=' + self.pipename()
        server.listen(addr)
        server.max_connections = 2
        clients = []
        exc = None
        try:
            for i in range(4):
                client = DbusClient()
                client.connect(addr)
                clients.append(client)
        except Exception as e:
            exc = e
        self.assertIsInstance(exc, TransportError)
        self.assertEqual(len(server.connections), server.max_connections)
        for client in clients:
            client.close()
        server.close()


class TestNativeDbus(UnitTest):

    def setUp(self):
        if not os.environ.get('DBUS_SESSION_BUS_ADDRESS'):
            raise SkipTest('D-BUS session bus not available')
        super(TestNativeDbus, self).setUp()

    def test_get_unique_name(self):
        # Ensure that get_unique_name() works
        client = DbusClient()
        client.connect('session')
        unique_name = client.get_unique_name()
        self.assertIsInstance(unique_name, six.text_type)
        self.assertTrue(unique_name.startswith(':'))
        client.close()

    def test_call_listnames(self):
        # Call the ListNames() bus method and ensure the results are a list of
        # strings.
        client = DbusClient()
        client.connect('session')
        result = client.call_method('org.freedesktop.DBus', '/org/freedesktop/DBus',
                                    'org.freedesktop.DBus', 'ListNames')
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 1)
        names = result[0]
        self.assertIsInstance(names, list)
        self.assertGreater(len(names), 0)
        for name in names:
            self.assertIsInstance(name, six.text_type)
        client.close()


if __name__ == '__main__':
    os.environ.setdefault('VERBOSE', '1')
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import fibers

from . import logging
from .hub import get_hub, switchpoint
from .sync import Event
from .errors import Cancelled, Timeout
from .callbacks import add_callback, remove_callback, run_callbacks

__all__ = ['current_fiber', 'Fiber', 'spawn']


def current_fiber():
    """Return the current fiber.

    Note: The root and hub fiber are "bare" :class:`fibers.Fiber` instances.
    Calling this method there returns the bare instance, not a
    :class:`gruvi.Fiber` instance.
    """
    return fibers.current()


class Fiber(fibers.Fiber):
    """An cooperatively scheduled execution context aka *green thread* aka
    *co-routine*."""

    # This class is a very thin layer on top of fibers.Fiber. It adds a start()
    # method that schedules a switch via the hub. It also enforces that only
    # the hub may call switch().
    #
    # All user created fibers should use this interface. The only fibers in a
    # Gruvi application that use the "raw" interface from the fibers package
    # are the root fiber and the Hub.

    __slots__ = ('_name', 'context', '_target', '_log', '_done', '_callbacks')

    def __init__(self, target, args=(), kwargs={}, name=None, hub=None):
        """
        The *target* argument is the main function of the fiber. It must be a
        Python callable. The *args* and *kwargs* specify its arguments and
        keyword arguments, respectively.

        The *name* argument specifies the fiber name. This is purely a
        diagnositic tool is used e.g. in log messages.

        The *hub* argument can be used to override the hub that will be used to
        schedule this fiber. This argument is used by the unit tests and should
        not by needed.
        """
        self._hub = hub or get_hub()
        super(Fiber, self).__init__(self.run, args, kwargs, self._hub)
        if name is None:
            fid = self._hub.data.setdefault('gruvi:next_fiber', 1)
            name = 'Fiber-{}'.format(fid)
            self._hub.data['gruvi:next_fiber'] += 1
        self._name = name
        self._target = target
        self._log = logging.get_logger()
        self._done = Event()
        self._callbacks = None

    @property
    def name(self):
        """The fiber's name."""
        return self._name

    @property
    def alive(self):
        """Whether the fiber is alive."""
        return self.is_alive()

    def start(self):
        """Schedule the fiber to be started in the next iteration of the
        event loop."""
        target = getattr(self._target, '__qualname__', self._target.__name__)
        self._log.debug('starting fiber {}, target {}', self.name, target)
        self._hub.run_callback(self.switch)

    def switch(self, value=None):
        # Only the hub may call this.
        if self.current() is not self._hub:
            raise RuntimeError('only the Hub may switch() to a fiber')
        if not self.is_alive():
            self._log.warning('attempt to switch to a dead Fiber')
            return
        return super(Fiber, self).switch(value)

    def throw(self, typ, val=None, tb=None):
        # Only the hub may call this.
        if self.current() is not self._hub:
            raise RuntimeError('only the Hub may throw() into a fiber')
        return super(Fiber, self).throw(typ, val, tb)

    def cancel(self, message=None):
        """Schedule the fiber to be cancelled in the next iteration of the
        event loop.

        Cancellation works by throwing a :class:`~gruvi.Cancelled` exception
        into the fiber. If *message* is provided, it will be set as the value
        of the exception.
        """
        if not self.is_alive():
            return
        if message is None:
            message = 'cancelled by Fiber.cancel()'
        self._hub.run_callback(self.throw, Cancelled, Cancelled(message))

    @switchpoint
    def join(self, timeout=None):
        """Wait until the fiber completes."""
        if not self._done.wait(timeout):
            raise Timeout('timeout waiting for fiber to exit')

    def run(self, *args, **kwargs):
        # Target of the first :meth:`switch()` call.
        if self.current() is not self:
            raise RuntimeError('run() may only be called from self')
        try:
            self._target(*args, **kwargs)
        except Cancelled as e:
            self._log.debug('fiber was cancelled ({!s})', e)
        except BaseException:
            self._log.exception('uncaught exception in fiber')
        self._done.set()
        run_callbacks(self)

    # Support wait()

    def add_done_callback(self, callback, *args):
        if self._done.is_set():
            callback(*args)
            return
        return add_callback(self, callback, args)

    def remove_done_callback(self, handle):
        remove_callback(self, handle)


def spawn(func, *args, **kwargs):
    """Spawn a new fiber.

    A new :class:`Fiber` is created with main function *func* and positional
    arguments *args*. The keyword arguments are passed to the :class:`Fiber`
    constructor, not to the main function. The fiber is then scheduled to start
    by calling its :meth:`~Fiber.start` method.

    The fiber instance is returned.
    """
    fiber = Fiber(func, args, **kwargs)
    fiber.start()
    return fiber



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import os
import json
import unittest
import six

import gruvi
from gruvi import jsonrpc
from gruvi.jsonrpc import JsonRpcError, JsonRpcVersion
from gruvi.jsonrpc import JsonRpcProtocol, JsonRpcClient, JsonRpcServer
from gruvi.jsonrpc_ffi import ffi as _ffi, lib as _lib
from gruvi.transports import TransportError
from support import UnitTest, MockTransport


_keepalive = None

def set_buffer(ctx, buf):
    global _keepalive  # See note in JsonRpcProtocol
    _keepalive = ctx.buf = _ffi.from_buffer(buf)
    ctx.buflen = len(buf)
    ctx.offset = 0

def split_string(s):
    ctx = _ffi.new('struct split_context *')
    set_buffer(ctx, s)
    _lib.json_split(ctx)
    return ctx


JsonRpcProtocol.default_version = '1.0'


class TestJsonSplitter(UnitTest):

    def test_simple(self):
        r = b'{ "foo": "bar" }'
        ctx = split_string(r)
        self.assertEqual(ctx.error, 0)
        self.assertEqual(ctx.offset, len(r))

    def test_leading_whitespace(self):
        r = b' { "foo": "bar" }'
        ctx = split_string(r)
        self.assertEqual(ctx.error, 0)
        self.assertEqual(ctx.offset, len(r))
        r = b' \t\n{ "foo": "bar" }'
        ctx = split_string(r)
        self.assertEqual(ctx.error, 0)
        self.assertEqual(ctx.offset, len(r))

    def test_trailing_whitespace(self):
        r = b'{ "foo": "bar" } '
        ctx = split_string(r)
        self.assertEqual(ctx.error, 0)
        self.assertEqual(ctx.offset, len(r)-1)
        error = _lib.json_split(ctx)
        self.assertEqual(error, ctx.error) == _lib.INCOMPLETE
        self.assertEqual(ctx.offset, len(r))

    def test_brace_in_string(self):
        r = b'{ "foo": "b{r" }'
        ctx = split_string(r)
        self.assertEqual(ctx.error, 0)
        self.assertEqual(ctx.offset, len(r))
        r = b'{ "foo": "b}r" }'
        ctx = split_string(r)
        self.assertEqual(ctx.error, 0)
        self.assertEqual(ctx.offset, len(r))

    def test_string_escape(self):
        r = b'{ "foo": "b\\"}" }'
        ctx = split_string(r)
        self.assertEqual(ctx.error, 0)
        self.assertEqual(ctx.offset, len(r))

    def test_error(self):
        r = b' x { "foo": "bar" }'
        ctx = split_string(r)
        self.assertEqual(ctx.error, _lib.ERROR)
        self.assertEqual(ctx.offset, 1)
        r = b'[ { "foo": "bar" } ]'
        ctx = split_string(r)
        self.assertEqual(ctx.error, _lib.ERROR)
        self.assertEqual(ctx.offset, 0)

    def test_multiple(self):
        r = b'{ "foo": "bar" } { "baz": "qux" }'
        ctx = split_string(r)
        self.assertEqual(ctx.error, 0)
        self.assertEqual(ctx.offset, 16)
        error = _lib.json_split(ctx)
        self.assertEqual(error, ctx.error) == 0
        self.assertEqual(ctx.offset, len(r))

    def test_incremental(self):
        r = b'{ "foo": "bar" }'
        ctx = _ffi.new('struct split_context *')
        for i in range(len(r)-1):
            set_buffer(ctx, r[i:i+1])
            error = _lib.json_split(ctx)
            self.assertEqual(error, ctx.error) == _lib.INCOMPLETE
            self.assertEqual(ctx.offset, 1)
        set_buffer(ctx, r[-1:])
        error = _lib.json_split(ctx)
        self.assertEqual(error, ctx.error) == 0
        self.assertEqual(ctx.offset, 1)


class TestJsonRpcV1(UnitTest):

    def setUp(self):
        super(TestJsonRpcV1, self).setUp()
        self.version = JsonRpcVersion.create('1.0')

    def test_check_request(self):
        v = self.version
        msg = {'id': 1, 'method': 'foo', 'params': []}
        self.assertEqual(v.check_message(msg), jsonrpc.REQUEST)
        msg = {'id': None, 'method': 'foo', 'params': []}
        self.assertEqual(v.check_message(msg), jsonrpc.REQUEST)

    def test_check_request_missing_id(self):
        v = self.version
        msg = {'method': 'foo', 'params': []}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_request_missing_method(self):
        v = self.version
        msg = {'id': 1, 'params': []}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_request_illegal_method(self):
        v = self.version
        msg = {'id': 1, 'method': None, 'params': []}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'id': 1, 'method': 1, 'params': []}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'id': 1, 'method': {}, 'params': []}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'id': 1, 'method': [], 'params': []}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'id': 1, 'method': [1], 'params': []}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_request_missing_params(self):
        v = self.version
        msg = {'id': 1, 'method': 'foo'}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_request_illegal_params(self):
        v = self.version
        msg = {'id': 1, 'method': 'foo', 'params': None}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'id': 1, 'method': 'foo', 'params': 1}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'id': 1, 'method': 'foo', 'params': 'foo'}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'id': 1, 'method': 'foo', 'params': {}}

    def test_check_request_extraneous_fields(self):
        v = self.version
        msg = {'id': 1, 'method': 'foo', 'params': [], 'bar': 'baz'}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_response(self):
        v = self.version
        msg = {'id': 1, 'result': 'foo', 'error': None}
        self.assertEqual(v.check_message(msg), jsonrpc.RESPONSE)

    def test_check_response_null_result(self):
        v = self.version
        msg = {'id': 1, 'result': None, 'error': None}
        self.assertEqual(v.check_message(msg), jsonrpc.RESPONSE)

    def test_check_response_error(self):
        v = self.version
        msg = {'id': 1, 'result': None, 'error': {'code': 1}}
        self.assertEqual(v.check_message(msg), jsonrpc.RESPONSE)

    def test_check_response_missing_id(self):
        v = self.version
        msg = {'result': 'foo', 'error': None}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_response_missing_result(self):
        v = self.version
        msg = {'id': 1, 'error': None}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_response_missing_error(self):
        v = self.version
        msg = {'id': 1, 'result': None}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_response_illegal_error(self):
        v = self.version
        msg = {'id': 1, 'result': None, 'error': 1}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'id': 1, 'result': None, 'error': 'foo'}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'id': 1, 'result': None, 'error': []}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_response_result_error_both_set(self):
        v = self.version
        msg = {'id': 1, 'result': 1, 'error': 0}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_response_extraneous_fields(self):
        v = self.version
        msg = {'id': 1, 'result': 1, 'error': None, 'bar': 'baz'}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_create_request(self):
        v = self.version
        msg = v.create_request('foo', [])
        self.assertIsInstance(msg['id'], six.string_types)
        self.assertEqual(msg['method'], 'foo')
        self.assertEqual(msg['params'], [])
        self.assertEqual(len(msg), 3)

    def test_create_request_notification(self):
        v = self.version
        msg = v.create_request('foo', [], notification=True)
        self.assertIsNone(msg['id'])
        self.assertEqual(msg['method'], 'foo')
        self.assertEqual(msg['params'], [])
        self.assertEqual(len(msg), 3)

    def test_create_response(self):
        v = self.version
        req = {'id': 'gruvi.0'}
        msg = v.create_response(req, 1)
        self.assertEqual(msg['id'], req['id'])
        self.assertEqual(msg['result'], 1)
        self.assertIsNone(msg['error'])
        self.assertEqual(len(msg), 3)

    def test_create_response_null_result(self):
        v = self.version
        req = {'id': 'gruvi.0'}
        msg = v.create_response(req, None)
        self.assertEqual(msg['id'], req['id'])
        self.assertIsNone(msg['result'])
        self.assertIsNone(msg['error'])
        self.assertEqual(len(msg), 3)

    def test_create_response_error(self):
        v = self.version
        req = {'id': 'gruvi.0'}
        msg = v.create_response(req, error={'code': 1})
        self.assertEqual(msg['id'], req['id'])
        self.assertIsNone(msg['result'])
        self.assertEqual(msg['error'], {'code': 1})
        self.assertEqual(len(msg), 3)


class TestJsonRpcV2(UnitTest):

    def setUp(self):
        super(TestJsonRpcV2, self).setUp()
        self.version = JsonRpcVersion.create('2.0')

    def test_check_request(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'id': 1, 'method': 'foo', 'params': []}
        self.assertEqual(v.check_message(msg), jsonrpc.REQUEST)
        msg = {'jsonrpc': '2.0', 'id': 1, 'method': 'foo', 'params': {}}
        self.assertEqual(v.check_message(msg), jsonrpc.REQUEST)

    def test_check_request_notification(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'method': 'foo', 'params': []}
        self.assertEqual(v.check_message(msg), jsonrpc.REQUEST)
        msg = {'jsonrpc': '2.0', 'method': 'foo', 'params': {}}
        self.assertEqual(v.check_message(msg), jsonrpc.REQUEST)

    def test_check_request_missing_version(self):
        v = self.version
        msg = {'id': 1, 'method': 'foo', 'params': []}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_request_missing_method(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'id': 1, 'params': []}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_request_illegal_method(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'id': 1, 'method': None, 'params': []}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'jsonrpc': '2.0', 'id': 1, 'method': 1, 'params': []}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'jsonrpc': '2.0', 'id': 1, 'method': {}, 'params': []}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'jsonrpc': '2.0', 'id': 1, 'method': [], 'params': []}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'jsonrpc': '2.0', 'id': 1, 'method': [1], 'params': []}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_request_missing_params(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'id': 1, 'method': 'foo'}
        self.assertEqual(v.check_message(msg), jsonrpc.REQUEST)

    def test_check_request_illegal_params(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'id': 1, 'method': 'foo', 'params': None}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'jsonrpc': '2.0', 'id': 1, 'method': 'foo', 'params': 1}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'jsonrpc': '2.0', 'id': 1, 'method': 'foo', 'params': 'foo'}

    def test_check_request_extraneous_fields(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'id': 1, 'method': 'foo', 'params': [], 'bar': 'baz'}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_response(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'id': 1, 'result': 'foo'}
        self.assertEqual(v.check_message(msg), jsonrpc.RESPONSE)

    def test_check_response_null_result(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'id': 1, 'result': None}
        self.assertEqual(v.check_message(msg), jsonrpc.RESPONSE)

    def test_check_response_error(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'id': 1, 'error': {'code': 1}}
        self.assertEqual(v.check_message(msg), jsonrpc.RESPONSE)

    def test_check_response_missing_id(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'result': 'foo'}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_response_null_id(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'id': None, 'result': 'foo'}
        self.assertEqual(v.check_message(msg), jsonrpc.RESPONSE)

    def test_check_response_error_missing_id(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'error': {'code': 10}}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_response_error_null_id(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'id': None, 'error': {'code': 1}}
        self.assertEqual(v.check_message(msg), jsonrpc.RESPONSE)

    def test_check_response_missing_result_and_error(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'id': 1}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_response_illegal_error(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'id': 1, 'error': 1}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'jsonrpc': '2.0', 'id': 1, 'error': 'foo'}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'jsonrpc': '2.0', 'id': 1, 'error': []}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_response_result_error_both_present(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'id': 1, 'result': None, 'error': None}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'jsonrpc': '2.0', 'id': 1, 'result': 1, 'error': None}
        self.assertRaises(ValueError, v.check_message, msg)
        msg = {'jsonrpc': '2.0', 'id': 1, 'result': None, 'error': {'code': 10}}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_check_response_extraneous_fields(self):
        v = self.version
        msg = {'jsonrpc': '2.0', 'id': 1, 'result': 1, 'error': None, 'bar': 'baz'}
        self.assertRaises(ValueError, v.check_message, msg)

    def test_create_request(self):
        v = self.version
        msg = v.create_request('foo', [])
        self.assertEqual(msg['jsonrpc'], '2.0')
        self.assertIsInstance(msg['id'], six.string_types)
        self.assertEqual(msg['method'], 'foo')
        self.assertEqual(msg['params'], [])
        self.assertEqual(len(msg), 4)

    def test_create_request_notification(self):
        v = self.version
        msg = v.create_request('foo', [], notification=True)
        self.assertEqual(msg['jsonrpc'], '2.0')
        self.assertNotIn('id', msg)
        self.assertEqual(msg['method'], 'foo')
        self.assertEqual(msg['params'], [])
        self.assertEqual(len(msg), 3)

    def test_create_response(self):
        v = self.version
        req = {'id': 'gruvi.0'}
        msg = v.create_response(req, 1)
        self.assertEqual(msg['jsonrpc'], '2.0')
        self.assertEqual(msg['id'], req['id'])
        self.assertEqual(msg['result'], 1)
        self.assertNotIn('error', msg)
        self.assertEqual(len(msg), 3)

    def test_create_response_null_result(self):
        v = self.version
        req = {'id': 'gruvi.0'}
        msg = v.create_response(req, None)
        self.assertEqual(msg['jsonrpc'], '2.0')
        self.assertEqual(msg['id'], req['id'])
        self.assertIsNone(msg['result'])
        self.assertNotIn('error', msg)
        self.assertEqual(len(msg), 3)

    def test_create_response_error(self):
        v = self.version
        req = {'id': 'gruvi.0'}
        msg = v.create_response(req, error={'code': 1})
        self.assertEqual(msg['jsonrpc'], '2.0')
        self.assertEqual(msg['id'], req['id'])
        self.assertNotIn('result', msg)
        self.assertEqual(msg['error'], {'code': 1})
        self.assertEqual(len(msg), 3)


class TestJsonRpcProtocol(UnitTest):

    def setUp(self):
        super(TestJsonRpcProtocol, self).setUp()
        self.transport = MockTransport()
        self.protocol = JsonRpcProtocol(self.message_handler)
        self.transport.start(self.protocol)
        self.messages = []
        self.protocols = []

    def message_handler(self, message, transport, protocol):
        self.messages.append(message)
        self.protocols.append(protocol)

    def get_messages(self):
        # run dispatcher thread so that it calls our message handler
        gruvi.sleep(0)
        return self.messages

    def test_simple(self):
        m = b'{ "id": "1", "method": "foo", "params": [] }'
        proto = self.protocol
        proto.data_received(m)
        mm = self.get_messages()
        self.assertEqual(len(mm), 1)
        self.assertIsInstance(mm[0], dict)
        self.assertEqual(mm[0], {'id': '1', 'method': 'foo', 'params': []})
        pp = self.protocols
        self.assertEqual(len(pp), 1)
        self.assertIs(pp[0], proto)

    def test_multiple(self):
        m = b'{ "id": "1", "method": "foo", "params": [] }' \
            b'{ "id": "2", "method": "bar", "params": [] }'
        proto = self.protocol
        proto.data_received(m)
        mm = self.get_messages()
        self.assertEqual(len(mm), 2)
        self.assertEqual(mm[0], {'id': '1', 'method': 'foo', 'params': []})
        self.assertEqual(mm[1], {'id': '2', 'method': 'bar', 'params': []})
        pp = self.protocols
        self.assertEqual(len(pp), 2)
        self.assertIs(pp[0], proto)
        self.assertIs(pp[1], proto)

    def test_whitespace(self):
        m = b'  { "id": "1", "method": "foo", "params": [] }' \
            b'  { "id": "2", "method": "bar", "params": [] }'
        proto = self.protocol
        proto.data_received(m)
        mm = self.get_messages()
        self.assertEqual(len(mm), 2)
        self.assertEqual(mm[0], {'id': '1', 'method': 'foo', 'params': []})
        self.assertEqual(mm[1], {'id': '2', 'method': 'bar', 'params': []})

    def test_incremental(self):
        m = b'{ "id": "1", "method": "foo", "params": [] }'
        proto = self.protocol
        for i in range(len(m)-1):
            proto.data_received(m[i:i+1])
            self.assertEqual(self.get_messages(), [])
        proto.data_received(m[-1:])
        mm = self.get_messages()
        self.assertEqual(len(mm), 1)
        self.assertEqual(mm[0], {'id': '1', 'method': 'foo', "params": []})

    def test_framing_error(self):
        m = b'xxx'
        proto = self.protocol
        proto.data_received(m)
        self.assertEqual(self.get_messages(), [])
        self.assertIsInstance(proto._error, JsonRpcError)

    def test_encoding_error(self):
        m = b'{ xxx\xff }'
        proto = self.protocol
        proto.data_received(m)
        self.assertEqual(self.get_messages(), [])
        self.assertIsInstance(proto._error, JsonRpcError)

    def test_illegal_json(self):
        m = b'{ "xxxx" }'
        proto = self.protocol
        proto.data_received(m)
        self.assertEqual(self.get_messages(), [])
        self.assertIsInstance(proto._error, JsonRpcError)

    def test_illegal_jsonrpc(self):
        m = b'{ "xxxx": "yyyy" }'
        proto = self.protocol
        proto.data_received(m)
        self.assertEqual(self.get_messages(), [])
        self.assertIsInstance(proto._error, JsonRpcError)

    def test_maximum_message_size_exceeded(self):
        proto = self.protocol
        proto.max_message_size = 100
        message = {'id': 1, 'method': 'foo', 'params': ['x'*100]}
        message = json.dumps(message).encode('utf8')
        self.assertGreater(len(message), proto.max_message_size)
        proto.data_received(message)
        self.assertEqual(self.get_messages(), [])
        self.assertIsInstance(proto._error, JsonRpcError)

    def test_flow_control(self):
        # Write more messages than the protocol is willing to pipeline. Flow
        # control should kick in and alternate scheduling of the producer and
        # the consumer.
        proto, trans = self.protocol, self.transport
        self.assertTrue(trans._reading)
        proto.max_pipeline_size = 10
        message = b'{ "id": 1, "method": "foo", "params": [] }'
        interrupted = 0
        for i in range(1000):
            proto.data_received(message)
            if not trans._reading:
                interrupted += 1
                gruvi.sleep(0)  # run dispatcher
            self.assertTrue(trans._reading)
        mm = self.get_messages()
        self.assertEqual(len(mm), 1000)
        self.assertEqual(interrupted, 100)
        message = json.loads(message.decode('utf8'))
        for m in mm:
            self.assertEqual(m, message)


def echo_app(message, transport, protocol):
    if message.get('method') != 'echo':
        protocol.send_response(message, error={'code': jsonrpc.METHOD_NOT_FOUND})
    else:
        protocol.send_response(message, message['params'])

def reflect_app(message, transport, protocol):
    if message.get('method') != 'echo':
        return
    value = protocol.call_method('echo', *message['params'])
    protocol.send_response(message, value)

def notification_app():
    notifications = []
    def application(message, transport, protocol):
        if message.get('id') is None:
            notifications.append((message['method'], message['params']))
        elif message['method'] == 'get_notifications':
            protocol.send_response(message, notifications)
    return application


class TestJsonRpc(UnitTest):

    def test_errno(self):
        code = jsonrpc.SERVER_ERROR
        self.assertIsInstance(code, int)
        name = jsonrpc.errorcode[code]
        self.assertIsInstance(name, str)
        self.assertEqual(getattr(jsonrpc, name), code)
        desc = jsonrpc.strerror(code)
        self.assertIsInstance(desc, str)

    def test_call_method_tcp(self):
        server = JsonRpcServer(echo_app)
        server.listen(('localhost', 0))
        addr = server.addresses[0]
        client = JsonRpcClient()
        client.connect(addr)
        result = client.call_method('echo', 'foo')
        self.assertEqual(result, ['foo'])
        server.close()
        client.close()

    def test_call_method_pipe(self):
        server = JsonRpcServer(echo_app)
        server.listen(self.pipename(abstract=True))
        addr = server.addresses[0]
        client = JsonRpcClient()
        client.connect(addr)
        result = client.call_method('echo', 'foo')
        self.assertEqual(result, ['foo'])
        server.close()
        client.close()

    def test_call_method_ssl(self):
        server = JsonRpcServer(echo_app)
        context = self.get_ssl_context()
        server.listen(('localhost', 0), ssl=context)
        addr = server.addresses[0]
        client = JsonRpcClient()
        client.connect(addr, ssl=context)
        result = client.call_method('echo', 'foo')
        self.assertEqual(result, ['foo'])
        server.close()
        client.close()

    def test_call_method_no_args(self):
        server = JsonRpcServer(echo_app)
        server.listen(('127.0.0.1', 0))
        addr = server.addresses[0]
        client = JsonRpcClient()
        client.connect(addr)
        result = client.call_method('echo')
        self.assertEqual(result, [])
        server.close()
        client.close()

    def test_call_method_multiple_args(self):
        server = JsonRpcServer(echo_app)
        server.listen(('127.0.0.1', 0))
        addr = server.addresses[0]
        client = JsonRpcClient()
        client.connect(addr)
        result = client.call_method('echo', 'foo', 'bar')
        self.assertEqual(result, ['foo', 'bar'])
        server.close()
        client.close()

    def test_call_method_error(self):
        server = JsonRpcServer(echo_app)
        server.listen(('127.0.0.1', 0))
        addr = server.addresses[0]
        client = JsonRpcClient()
        client.connect(addr)
        exc = self.assertRaises(JsonRpcError, client.call_method, 'echo2')
        self.assertIsInstance(exc, JsonRpcError)
        self.assertIsInstance(exc.error, dict)
        self.assertEqual(exc.error['code'], jsonrpc.METHOD_NOT_FOUND)
        server.close()
        client.close()

    def test_send_notification(self):
        server = JsonRpcServer(notification_app())
        server.listen(('127.0.0.1', 0))
        addr = server.addresses[0]
        client = JsonRpcClient()
        client.connect(addr)
        client.send_notification('notify_foo', 'foo')
        notifications = client.call_method('get_notifications')
        self.assertEqual(notifications, [['notify_foo', ['foo']]])
        server.close()
        client.close()

    def test_call_method_ping_pong(self):
        server = JsonRpcServer(reflect_app)
        server.listen(('127.0.0.1', 0))
        addr = server.addresses[0]
        client = JsonRpcClient(echo_app)
        client.connect(addr)
        result = client.call_method('echo', 'foo')
        self.assertEqual(result, ['foo'])
        server.close()
        client.close()

    def test_send_evil(self):
        server = JsonRpcServer(echo_app)
        server.listen(('127.0.0.1', 0))
        addr = server.addresses[0]
        client = JsonRpcClient()
        client.connect(addr)
        exc = None
        try:
            chunk = b'{' * 1024
            while True:
                client.transport.write(chunk)
                gruvi.sleep(0)
        except Exception as e:
            exc = e
        self.assertIsInstance(exc, TransportError)
        server.close()
        client.close()

    def test_send_whitespace(self):
        server = JsonRpcServer(echo_app)
        server.listen(('127.0.0.1', 0))
        addr = server.addresses[0]
        client = JsonRpcClient()
        client.connect(addr)
        exc = None
        try:
            chunk = b' ' * 1024
            while True:
                client.transport.write(chunk)
                gruvi.sleep(0)
        except Exception as e:
            exc = e
        self.assertIsInstance(exc, TransportError)
        server.close()
        client.close()

    def test_send_random(self):
        server = JsonRpcServer(echo_app)
        server.listen(('127.0.0.1', 0))
        addr = server.addresses[0]
        client = JsonRpcClient()
        client.connect(addr)
        exc = None
        try:
            while True:
                chunk = os.urandom(1024)
                client.transport.write(chunk)
                gruvi.sleep(0)
        except Exception as e:
            exc = e
        self.assertIsInstance(exc, TransportError)
        server.close()
        client.close()

    def test_connection_limit(self):
        server = JsonRpcServer(echo_app)
        server.listen(('127.0.0.1', 0))
        addr = server.addresses[0]
        server.max_connections = 2
        clients = []
        exc = None
        try:
            for i in range(3):
                client = JsonRpcClient(timeout=2)
                client.connect(addr)
                client.call_method('echo')
                clients.append(client)
        except Exception as e:
            exc = e
        self.assertIsInstance(exc, TransportError)
        self.assertEqual(len(server.connections), server.max_connections)
        for client in clients:
            client.close()
        server.close()


if __name__ == '__main__':
    unittest.main()
